# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#    
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as 
# published by the Free Software Foundation, either version 3 of the License, 
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.
"""
Copyright 2014-2019 Anthony Larcher and Sylvain Meignier

    :mod:`iv_scoring` provides methods to compare i-vectors
"""
import copy
import logging
import numpy
import scipy
from sidekit.bosaris import Ndx
from sidekit.bosaris import Scores
from sidekit.statserver import StatServer

import sys
if sys.version_info.major > 2:
    from functools import reduce


__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2014-2019 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


def _check_missing_model(enroll, test, ndx):
    # Remove missing models and test segments
    clean_ndx = ndx.filter(enroll.modelset, test.segset, True)

    # Align StatServers to match the clean_ndx
    enroll.align_models(clean_ndx.modelset)
    test.align_segments(clean_ndx.segset)

    return clean_ndx


def cosine_scoring(enroll, test, ndx, wccn=None, check_missing=True):
    """Compute the cosine similarities between to sets of vectors. The list of 
    trials to perform is given in an Ndx object.
    
    :param enroll: a StatServer in which stat1 are i-vectors
    :param test: a StatServer in which stat1 are i-vectors
    :param ndx: an Ndx object defining the list of trials to perform
    :param wccn: numpy.ndarray, if provided, the i-vectors are normalized by using a Within Class Covariance Matrix
    :param check_missing: boolean, if True, check that all models and segments exist
    
    :return: a score object
    """
    assert isinstance(enroll, StatServer), 'First parameter should be a StatServer'
    assert isinstance(test, StatServer), 'Second parameter should be a StatServer'
    assert isinstance(ndx, Ndx), 'Third parameter should be an Ndx'
    enroll_copy = copy.deepcopy(enroll)
    test_copy = copy.deepcopy(test)

    # If models are not unique, compute the mean per model, display a warning
    #if not numpy.unique(enroll_copy.modelset).shape == enroll_copy.modelset.shape:
    #    logging.warning("Enrollment models are not unique, average i-vectors")
    #    enroll_copy = enroll_copy.mean_stat_per_model()

    # Remove missing models and test segments
    if check_missing:
        clean_ndx = _check_missing_model(enroll_copy, test_copy, ndx)
    else:
        clean_ndx = ndx

    if wccn is not None:
        enroll_copy.rotate_stat1(wccn)
        if enroll_copy != test_copy:
            test_copy.rotate_stat1(wccn)

    # Cosine scoring
    enroll_copy.norm_stat1()
    if enroll_copy != test_copy:
        test_copy.norm_stat1()
    s = numpy.dot(enroll_copy.stat1, test_copy.stat1.transpose())

    score = Scores()
    score.scoremat = s
    score.modelset = clean_ndx.modelset
    score.segset = clean_ndx.segset
    score.scoremask = clean_ndx.trialmask
    return score


def mahalanobis_scoring(enroll, test, ndx, m, check_missing=True):
    """Compute the mahalanobis distance between to sets of vectors. The list of 
    trials to perform is given in an Ndx object.
    
    :param enroll: a StatServer in which stat1 are i-vectors
    :param test: a StatServer in which stat1 are i-vectors
    :param ndx: an Ndx object defining the list of trials to perform
    :param m: mahalanobis matrix as a ndarray
    :param check_missing: boolean, default is True, set to False not to check missing models
    
    :return: a score object
    """
    assert isinstance(enroll, StatServer), 'First parameter should be a StatServer'
    assert isinstance(test, StatServer), 'Second parameter should be a StatServer'
    assert isinstance(ndx, Ndx), 'Third parameter should be an Ndx'
    assert enroll.stat1.shape[1] == test.stat1.shape[1], 'I-vectors dimension mismatch'
    assert enroll.stat1.shape[1] == m.shape[0], 'I-vectors and Mahalanobis matrix dimension mismatch'

    # If models are not unique, compute the mean per model, display a warning
    if not numpy.unique(enroll.modelset).shape == enroll.modelset.shape:
        logging.warning("Enrollment models are not unique, average i-vectors")
        enroll = enroll.mean_stat_per_model()

    # Remove missing models and test segments
    if check_missing:
        clean_ndx = _check_missing_model(enroll, test, ndx)
    else:
        clean_ndx = ndx

    # Mahalanobis scoring
    s = numpy.zeros((enroll.modelset.shape[0], test.segset.shape[0]))
    for i in range(enroll.modelset.shape[0]):
        diff = enroll.stat1[i, :] - test.stat1
        s[i, :] = -0.5 * numpy.sum(numpy.dot(diff, m) * diff, axis=1)

    score = Scores()
    score.scoremat = s
    score.modelset = clean_ndx.modelset
    score.segset = clean_ndx.segset
    score.scoremask = clean_ndx.trialmask
    return score


def two_covariance_scoring(enroll, test, ndx, W, B, check_missing=True):
    """Compute the 2-covariance scores between to sets of vectors. The list of 
    trials to perform is given in an Ndx object. Within and between class 
    co-variance matrices have to be pre-computed.
    
    :param enroll: a StatServer in which stat1 are i-vectors
    :param test: a StatServer in which stat1 are i-vectors
    :param ndx: an Ndx object defining the list of trials to perform
    :param W: the within-class co-variance matrix to consider
    :param B: the between-class co-variance matrix to consider
    :param check_missing: boolean, default is True, set to False not to check missing models
      
    :return: a score object
    """
    assert isinstance(enroll, StatServer), 'First parameter should be a directory'
    assert isinstance(test, StatServer), 'Second parameter should be a StatServer'
    assert isinstance(ndx, Ndx), 'Third parameter should be an Ndx'
    assert enroll.stat1.shape[1] == test.stat1.shape[1], 'I-vectors dimension mismatch'
    assert enroll.stat1.shape[1] == W.shape[0], 'I-vectors and co-variance matrix dimension mismatch'
    assert enroll.stat1.shape[1] == B.shape[0], 'I-vectors and co-variance matrix dimension mismatch'

    # If models are not unique, compute the mean per model, display a warning
    if not numpy.unique(enroll.modelset).shape == enroll.modelset.shape:
        logging.warning("Enrollment models are not unique, average i-vectors")
        enroll = enroll.mean_stat_per_model()

    # Remove missing models and test segments
    if check_missing:
        clean_ndx = _check_missing_model(enroll, test, ndx)
    else:
        clean_ndx = ndx

    # Two covariance scoring scoring
    S = numpy.zeros((enroll.modelset.shape[0], test.segset.shape[0]))
    iW = scipy.linalg.inv(W)
    iB = scipy.linalg.inv(B)

    G = reduce(numpy.dot, [iW, scipy.linalg.inv(iB + 2*iW), iW])
    H = reduce(numpy.dot, [iW, scipy.linalg.inv(iB + iW), iW])

    s2 = numpy.sum(numpy.dot(enroll.stat1, H) * enroll.stat1, axis=1)
    s3 = numpy.sum(numpy.dot(test.stat1, H) * test.stat1, axis=1)

    for ii in range(enroll.modelset.shape[0]):
        A = enroll.stat1[ii, :] + test.stat1
        s1 = numpy.sum(numpy.dot(A, G) * A, axis=1)
        S[ii, :] = s1 - s3 - s2[ii]

    score = Scores()
    score.scoremat = S
    score.modelset = clean_ndx.modelset
    score.segset = clean_ndx.segset
    score.scoremask = clean_ndx.trialmask
    return score


def PLDA_scoring(enroll,
                 test,
                 ndx,
                 mu,
                 F,
                 G,
                 Sigma,
                 test_uncertainty=None,
                 Vtrans=None,
                 p_known=0.0,
                 scaling_factor=1.,

                 full_model=False):
    """Compute the PLDA scores between to sets of vectors. The list of
    trials to perform is given in an Ndx object. PLDA matrices have to be
    pre-computed. i-vectors are supposed to be whitened before.

    Implements the appraoch described in [Lee13]_ including scoring
    for partially open-set identification

    :param enroll: a StatServer in which stat1 are i-vectors
    :param test: a StatServer in which stat1 are i-vectors
    :param ndx: an Ndx object defining the list of trials to perform
    :param mu: the mean vector of the PLDA gaussian
    :param F: the between-class co-variance matrix of the PLDA
    :param G: the within-class co-variance matrix of the PLDA
    :param Sigma: the residual covariance matrix
    :param p_known: probability of having a known speaker for open-set
        identification case (=1 for the verification task and =0 for the
        closed-set case)
    :param scaling_factor: scaling factor to be multiplied by the sufficient statistics
    :param full_model: boolean, set to True when using a complete PLDA model (including within class covariance matrix)

    :return: a score object
    """
    assert isinstance(enroll, StatServer), 'First parameter should be a StatServer'
    assert isinstance(test, StatServer), 'Second parameter should be a StatServer'
    assert isinstance(ndx, Ndx), 'Third parameter should be an Ndx'
    assert enroll.stat1.shape[1] == test.stat1.shape[1], 'I-vectors dimension mismatch'
    assert enroll.stat1.shape[1] == F.shape[0], 'I-vectors and co-variance matrix dimension mismatch'
    assert enroll.stat1.shape[1] == G.shape[0], 'I-vectors and co-variance matrix dimension mismatch'

    if not full_model:
        return fast_PLDA_scoring(enroll,
                                 test,
                                 ndx,
                                 mu,
                                 F,
                                 Sigma,
                                 test_uncertainty,
                                 Vtrans,
                                 p_known=p_known,
                                 scaling_factor=scaling_factor,
                                 check_missing=True)
    else:
        return full_PLDA_scoring(enroll, test, ndx, mu, F, G, Sigma, p_known=p_known, scaling_factor=scaling_factor)


def full_PLDA_scoring(enroll, test, ndx, mu, F, G, Sigma, p_known=0.0, scaling_factor=1., check_missing=True):
    """Compute PLDA scoring

    :param enroll: a StatServer in which stat1 are i-vectors
    :param test: a StatServer in which stat1 are i-vectors
    :param ndx: an Ndx object defining the list of trials to perform
    :param mu: the mean vector of the PLDA gaussian
    :param F: the between-class co-variance matrix of the PLDA
    :param G: the within-class co-variance matrix of the PLDA
    :param Sigma: the residual covariance matrix
    :param p_known: probability of having a known speaker for open-set
        identification case (=1 for the verification task and =0 for the
        closed-set case)
    :param check_missing: boolean, default is True, set to False not to check missing models

    """

    enroll_copy = copy.deepcopy(enroll)
    test_copy = copy.deepcopy(test)

    # If models are not unique, compute the mean per model, display a warning
    # if not numpy.unique(enroll_copy.modelset).shape == enroll_copy.modelset.shape:
    #    logging.warning("Enrollment models are not unique, average i-vectors")
    #    enroll_copy = enroll_copy.mean_stat_per_model()

    # Remove missing models and test segments
    if check_missing:
        clean_ndx = _check_missing_model(enroll_copy, test_copy, ndx)
    else:
        clean_ndx = ndx

    # Center the i-vectors around the PLDA mean
    enroll_copy.center_stat1(mu)
    test_copy.center_stat1(mu)

    # Compute temporary matrices
    invSigma = scipy.linalg.inv(Sigma)
    I_iv = numpy.eye(mu.shape[0], dtype='float')
    I_ch = numpy.eye(G.shape[1], dtype='float')
    I_spk = numpy.eye(F.shape[1], dtype='float')
    A = numpy.linalg.inv(G.T.dot(invSigma * scaling_factor).dot(G) + I_ch)  # keep numpy as interface are different
    B = F.T.dot(invSigma * scaling_factor).dot(I_iv - G.dot(A).dot(G.T).dot(invSigma * scaling_factor))
    K = B.dot(F)
    K1 = scipy.linalg.inv(K + I_spk)
    K2 = scipy.linalg.inv(2 * K + I_spk)

    # Compute the Gaussian distribution constant
    alpha1 = numpy.linalg.slogdet(K1)[1]
    alpha2 = numpy.linalg.slogdet(K2)[1]
    constant = alpha2 / 2.0 - alpha1

    # Compute verification scores
    score = Scores()
    score.scoremat = numpy.zeros(clean_ndx.trialmask.shape)
    score.modelset = clean_ndx.modelset
    score.segset = clean_ndx.segset
    score.scoremask = clean_ndx.trialmask

    # Project data in the space that maximizes the speaker separability
    test_tmp = B.dot(test_copy.stat1.T)
    enroll_tmp = B.dot(enroll_copy.stat1.T)

    # score qui ne dépend que du segment
    tmp1 = test_tmp.T.dot(K1)

    # Compute the part of the score that is only dependent on the test segment
    S1 = numpy.empty(test_copy.segset.shape[0])
    for seg_idx in range(test_copy.segset.shape[0]):
        S1[seg_idx] = tmp1[seg_idx, :].dot(test_tmp[:, seg_idx])/2.

    # Compute the part of the score that depends only on the model (S2) and on both model and test segment
    S2 = numpy.empty(enroll_copy.modelset.shape[0])

    for model_idx in range(enroll_copy.modelset.shape[0]):
        mod_plus_test_seg = test_tmp + numpy.atleast_2d(enroll_tmp[:, model_idx]).T
        tmp2 = mod_plus_test_seg.T.dot(K2)

        S2[model_idx] = enroll_tmp[:, model_idx].dot(K1).dot(enroll_tmp[:, model_idx])/2.
        score.scoremat[model_idx, :] = numpy.einsum("ij, ji->i", tmp2, mod_plus_test_seg)/2.

    score.scoremat += constant - (S1 + S2[:, numpy.newaxis])
    score.scoremat *= scaling_factor

    # Case of open-set identification, we compute the log-likelihood
    # by taking into account the probability of having a known impostor
    # or an out-of set class
    if p_known != 0:
        N = score.scoremat.shape[0]
        open_set_scores = numpy.empty(score.scoremat.shape)
        tmp = numpy.exp(score.scoremat)
        for ii in range(N):
            # open-set term
            open_set_scores[ii, :] = score.scoremat[ii, :] \
                - numpy.log(p_known * tmp[~(numpy.arange(N) == ii)].sum(axis=0) / (N - 1) + (1 - p_known))
        score.scoremat = open_set_scores

    return score

def fast_PLDA_scoring(enroll,
                      test,
                      ndx,
                      mu,
                      F,
                      Sigma,
                      test_uncertainty=None,
                      Vtrans=None,
                      p_known=0.0,
                      scaling_factor=1.,
                      check_missing=True):
    """Compute the PLDA scores between to sets of vectors. The list of
    trials to perform is given in an Ndx object. PLDA matrices have to be
    pre-computed. i-vectors are supposed to be whitened before.

    :param enroll: a StatServer in which stat1 are i-vectors
    :param test: a StatServer in which stat1 are i-vectors
    :param ndx: an Ndx object defining the list of trials to perform
    :param mu: the mean vector of the PLDA gaussian
    :param F: the between-class co-variance matrix of the PLDA
    :param Sigma: the residual covariance matrix
    :param p_known: probability of having a known speaker for open-set
        identification case (=1 for the verification task and =0 for the
        closed-set case)
    :param check_missing: boolean, if True, check that all models and segments exist

    :return: a score object
    """
    # assert isinstance(enroll, StatServer), 'First parameter should be a StatServer'
    # assert isinstance(test, StatServer), 'Second parameter should be a StatServer'
    # assert isinstance(ndx, Ndx), 'Third parameter should be an Ndx'
    # assert enroll.stat1.shape[1] == test.stat1.shape[1], 'I-vectors dimension mismatch'
    # assert enroll.stat1.shape[1] == F.shape[0], 'I-vectors and co-variance matrix dimension mismatch'
    # assert enroll.stat1.shape[1] == G.shape[0], 'I-vectors and co-variance matrix dimension mismatch'

    enroll_ctr = copy.deepcopy(enroll)
    test_ctr = copy.deepcopy(test)

    # If models are not unique, compute the mean per model, display a warning
    if not numpy.unique(enroll_ctr.modelset).shape == enroll_ctr.modelset.shape:
        logging.warning("Enrollment models are not unique, average i-vectors")
        enroll_ctr = enroll_ctr.mean_stat_per_model()

    # Remove missing models and test segments
    if check_missing:
        clean_ndx = _check_missing_model(enroll_ctr, test_ctr, ndx)
    else:
        clean_ndx = ndx

    # Center the i-vectors around the PLDA mean
    enroll_ctr.center_stat1(mu)
    test_ctr.center_stat1(mu)

    # If models are not unique, compute the mean per model, display a warning
    if not numpy.unique(enroll_ctr.modelset).shape == enroll_ctr.modelset.shape:
        logging.warning("Enrollment models are not unique, average i-vectors")
        enroll_ctr = enroll_ctr.mean_stat_per_model()

    # Compute constant component of the PLDA distribution
    invSigma = scipy.linalg.inv(Sigma)
    I_spk = numpy.eye(F.shape[1], dtype='float')

    K = F.T.dot(invSigma * scaling_factor).dot(F)
    K1 = scipy.linalg.inv(K + I_spk)
    K2 = scipy.linalg.inv(2 * K + I_spk)

    # Compute the Gaussian distribution constant
    alpha1 = numpy.linalg.slogdet(K1)[1]
    alpha2 = numpy.linalg.slogdet(K2)[1]
    plda_cst = alpha2 / 2.0 - alpha1

    # Compute intermediate matrices
    Sigma_ac = numpy.dot(F, F.T)
    Sigma_tot = Sigma_ac + Sigma
    Sigma_tot_inv = scipy.linalg.inv(Sigma_tot)

    Tmp = numpy.linalg.inv(Sigma_tot - Sigma_ac.dot(Sigma_tot_inv).dot(Sigma_ac))
    Phi = Sigma_tot_inv - Tmp
    Psi = Sigma_tot_inv.dot(Sigma_ac).dot(Tmp)

    # Compute the different parts of PLDA score
    model_part = 0.5 * numpy.einsum('ij, ji->i', enroll_ctr.stat1.dot(Phi), enroll_ctr.stat1.T)
    seg_part = 0.5 * numpy.einsum('ij, ji->i', test_ctr.stat1.dot(Phi), test_ctr.stat1.T)

    # Compute verification scores
    score = Scores()
    score.modelset = clean_ndx.modelset
    score.segset = clean_ndx.segset
    score.scoremask = clean_ndx.trialmask

    score.scoremat = model_part[:, numpy.newaxis] + seg_part + plda_cst
    score.scoremat += enroll_ctr.stat1.dot(Psi).dot(test_ctr.stat1.T)
    score.scoremat *= scaling_factor

    # Case of open-set identification, we compute the log-likelihood
    # by taking into account the probability of having a known impostor
    # or an out-of set class
    if p_known != 0:
        N = score.scoremat.shape[0]
        open_set_scores = numpy.empty(score.scoremat.shape)
        tmp = numpy.exp(score.scoremat)
        for ii in range(N):
            # open-set term
            open_set_scores[ii, :] = score.scoremat[ii, :] \
                - numpy.log(p_known * tmp[~(numpy.arange(N) == ii)].sum(axis=0) / (N - 1) + (1 - p_known))
        score.scoremat = open_set_scores

    return score


#IL FAUT RAJOUTER LA GESTION DES MULTI-SESSIONS (voir fonction PLDA_scoring_with_test_uncertainty_by_the_book de Themos)
#IMPLEMENTER LA VERSION "BY THE BOOK" pour ne pas utiliser la moyenne des i-vecteurs
def PLDA_scoring_uncertainty(enroll,
                             test,
                             ndx,
                             mu,
                             F,
                             Sigma,
                             p_known=0.0,
                             scaling_factor=1.,
                             test_uncertainty=None,
                             Vtrans=None,
                             check_missing=True):
    """

    :param enroll:
    :param test:
    :param ndx:
    :param mu:
    :param F:
    :param Sigma:
    :param p_known:
    :param scaling_factor:
    :param test_uncertainty:
    :param Vtrans:
    :param check_missing:
    :return:
    """
    assert isinstance(enroll, StatServer), 'First parameter should be a StatServer'
    assert isinstance(test, StatServer), 'Second parameter should be a StatServer'
    assert isinstance(ndx, Ndx), 'Third parameter should be an Ndx'
    assert enroll.stat1.shape[1] == test.stat1.shape[1], 'I-vectors dimension mismatch'
    assert enroll.stat1.shape[1] == F.shape[0], 'I-vectors and co-variance matrix dimension mismatch'
    assert enroll.stat1.shape[1] == G.shape[0], 'I-vectors and co-variance matrix dimension mismatch'

    enroll_ctr = copy.deepcopy(enroll)
    test_ctr = copy.deepcopy(test)

    # Remove missing models and test segments
    if check_missing:
        clean_ndx = _check_missing_model(enroll_ctr, test_ctr, ndx)
    else:
        clean_ndx = ndx

    # Center the i-vectors around the PLDA mean
    enroll_ctr.center_stat1(mu)
    test_ctr.center_stat1(mu)

    # Align StatServers to match the clean_ndx
    enroll_ctr.align_models_average(clean_ndx.modelset)
    test_ctr.align_segments(clean_ndx.segset)

    # Compute constant component of the PLDA distribution
    scoremat = numpy.zeros((enroll_ctr.stat1.shape[0], test_ctr.stat1.shape[0]), dtype='float')
    invSigma = scipy.linalg.inv(Sigma)

    K1 = scipy.linalg.inv(numpy.eye(F.shape[1]) + F.T.dot(invSigma * scaling_factor).dot(F))

    FK1Ft = F.dot(K1).dot(F.T)
    Xtilda_e = FK1Ft.dot(invSigma * scaling_factor).dot(enroll_ctr.stat1.T).T

    for t in range(test_ctr.stat1.shape[0]):
        xt = test_ctr.stat1[t,:]
        Pr = numpy.eye(F.shape[1]) - numpy.outer(test.stat1[t, :],test.stat1[t, :])
        Cunc = Pr.dot(Vtrans.transpose()).dot(numpy.diag(test_uncertainty[t, :]).dot(Vtrans)).dot(Pr)
        prec_den = scipy.linalg.inv(F.dot(F.T) + Sigma + Cunc)

        denom = -0.5 * xt.dot(prec_den).dot(xt) +0.5 * numpy.linalg.slogdet(prec_den)[1]
        prec_num = scipy.linalg.inv(FK1Ft+Sigma+Cunc)
        Xec = Xtilda_e - xt
        numer = -0.5 * numpy.einsum('ij, ji->i', Xec.dot(prec_num), Xec.T) + 0.5 * numpy.linalg.slogdet(prec_num)[1]
        scoremat[:, t] = scaling_factor * (numer - denom)

    # Compute verification scores
    score = Scores()
    score.modelset = clean_ndx.modelset
    score.segset = clean_ndx.segset
    score.scoremask = clean_ndx.trialmask

    score.scoremat = scoremat

    # Case of open-set identification, we compute the log-likelihood
    # by taking into account the probability of having a known impostor
    # or an out-of set class
    if p_known != 0:
        N = score.scoremat.shape[0]
        open_set_scores = numpy.empty(score.scoremat.shape)
        tmp = numpy.exp(score.scoremat)
        for ii in range(N):
            # open-set term
            open_set_scores[ii, :] = score.scoremat[ii, :] \
                - numpy.log(p_known * tmp[~(numpy.arange(N) == ii)].sum(axis=0) / (N - 1) + (1 - p_known))
        score.scoremat = open_set_scores

    return score



# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2020 Anthony Larcher
"""

import logging
import numpy
import torch
import pandas
import soundfile
import random
import h5py
import torch.optim as optim
import torch.multiprocessing as mp
from torchvision import transforms
from torch.utils.data import DataLoader
from pathlib import Path
from tqdm import tqdm
from torch.utils.data import Dataset


__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2015-2020 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reS'


"""
How to use:

vs = ds.Vox1Set("/lium/raid01_c/larcher/vox1_raw_wav_batches.h5", transform=transforms.Compose([PreEmphasis(),]))
vloader = DataLoader(vs, batch_size=32, shuffle=True, num_workers=5)
"""



def prepare_voxceleb1(vox1_root_dir, output_batch_file, seg_duration=4, samplerate=16000):

    # List wav files in VoxCeleb1
    vox1_wav_list = [str(f) for f in list(Path(vox1_root_dir).rglob("*.[wW][aA][vV]"))]
    vox1_df = pandas.DataFrame(columns=("database", "speaker_id", "file_id", "duration", "speaker_idx"))

    print("*** Collect information from VoxCeleb1 data ***")
    for fn in tqdm(vox1_wav_list):
        file_id = ('/').join(fn.split('/')[-2:]).split('.')[0]
        speaker_id = fn.split('/')[-3]
        _set = fn.split('/')[-5]

        # get the duration of the wav file
        data, _ = soundfile.read(fn)
        duration = data.shape[0]

        vox1_df = vox1_df.append(
            {"database": "vox1", "speaker_id": speaker_id, "file_id": file_id, "duration": duration, "speaker_idx": -1,
             "set": _set}, ignore_index=True)

    print("\n\n*** Create a single HDF5 file with all training data ***")

    # Create a HDF5 file and fill it with one 4s segment per session
    with h5py.File(output_batch_file, 'w') as fh:
        for index, row in tqdm(vox1_df.iterrows()):

            session_id = row['speaker_id'] + '/' + row['file_id']

            # Load the wav signal
            fn = '/'.join((vox1_root_dir, row['set'], 'wav', session_id)) + ".wav"
            data, samplerate = soundfile.read(fn, dtype='int16')
            _nb_samp = len(data)

            # Randomly select a segment of "duration" if it's long enough
            if _nb_samp > nb_samp:
                cut = numpy.random.randint(low = 0, high = _nb_samp - nb_samp)

                # Write the segment in the HDF5 file
                fh.create_dataset(session_id,
                                  data=data[cut:cut+nb_samp].astype('int16'),
                                  maxshape=(None,),
                                  fletcher32=True)


class PreEmphasis(object):
    """
    Perform pre-emphasis filtering on audio segment
    """
    def __init__(self, pre_emp_value=0.97):
        self.pre_emp_value = pre_emp_value

    def __call__(self, sample):
        data = numpy.asarray(sample[0][1:] - 0.97 * sample[0][:-1], dtype=numpy.float32)
        return data, sample[1]


class Vox1Set(Dataset):
    """
    Object creates a dataset for VoxCeleb
    """

    def __init__(self, voxceleb1_file, speaker_list=None, transform=None):
        """

        :param voxceleb1_file: HDF5 file containing data from Voxceleb1
        :param speaker_list:  list of speaker top use for training
        :param transform: list of transformation to apply
        """
        self.transform = transform
        self.voxceleb1_file = voxceleb1_file
        with h5py.File(voxceleb1_file, 'r') as fh:
            speaker_ids = list(fh.keys())

            # Filter speaker according to the input list
            if speaker_list is not None:
                speaker_ids = [l for l in speaker_ids if l in speaker_list]

            # Create a dictionary of speaker
            speaker_ids.sort()
            self.speaker_idx = {key: val for val, key in enumerate(speaker_ids)}

            l2 = []
            for l in speaker_ids:
                for k in list(fh[l].keys()):
                    l2.append('/'.join((l, k)))
            segment_list = []
            for l in l2:
                for k in list(fh[l].keys()):
                    segment_list.append('/'.join((l, k)))
            self.segment_list = segment_list
            self.len = len(self.segment_list)

    def __getitem__(self, index):
        with h5py.File(self.voxceleb1_file, 'r') as fh:
            speaker_idx = self.speaker_idx[self.segment_list[index].split('/')[0]]
            data_pcm16 = fh[self.segment_list[index]][()]
            data_float32 = data_pcm16.astype(numpy.float32) / 32768.
            if self.transform:
                data_float32, speaker_idx = self.transform((data_float32, speaker_idx))
            return data_float32, speaker_idx

    def __len__(self):
        return self.len


# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#    
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as 
# published by the Free Software Foundation, either version 3 of the License, 
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2020 Anthony Larcher and Sylvain Meignier
"""

from ctypes import *
from ctypes.util import find_library
import importlib
import logging
import numpy
import os
import sys



# Read environment variable if it exists
SIDEKIT_CONFIG={"libsvm":True,
                "mpi":False,
                "cuda":True
                }

if 'SIDEKIT' in os.environ:
    for cfg in os.environ['SIDEKIT'].split(","):
        k, val = cfg.split("=")
        if k == "libsvm":
           if val == "false":
                SIDEKIT_CONFIG["libsvm"] = False 
        elif k == "mpi":
            if val == "true":
               SIDEKIT_CONFIG["mpi"] = True
        if k == "cuda":
            if val == "true":
                SIDEKIT_CONFIG["cuda"] = True


PARALLEL_MODULE = 'multiprocessing'  # can be , threading, multiprocessing MPI is planned in the future
PARAM_TYPE = numpy.float32
STAT_TYPE = numpy.float64  # can be numpy.float32 to speed up the computation but can lead to numerical issuess

# Import bosaris-like classes
from .bosaris import IdMap
from .bosaris import Ndx
from .bosaris import Key
from .bosaris import Scores
from .bosaris import DetPlot
from .bosaris import effective_prior
from .bosaris import logit_effective_prior
from .bosaris import fast_minDCF

# Import classes
from .features_extractor import FeaturesExtractor
from .features_server import FeaturesServer
from .mixture import Mixture, vad_energy
from .statserver import StatServer
from .factor_analyser import FactorAnalyser

from .frontend.io import write_pcm
from .frontend.io import read_pcm
from .frontend.io import pcmu2lin
from .frontend.io import read_sph
from .frontend.io import write_label
from .frontend.io import read_label
from .frontend.io import read_spro4
from .frontend.io import read_audio
from .frontend.io import write_spro4
from .frontend.io import read_htk
from .frontend.io import write_htk

from .frontend.vad import vad_energy
from .frontend.vad import vad_snr
from .frontend.vad import label_fusion
from .frontend.vad import speech_enhancement

from .frontend.normfeat import cms
from .frontend.normfeat import cmvn
from .frontend.normfeat import stg
from .frontend.normfeat import rasta_filt

from .frontend.features import compute_delta
from .frontend.features import framing
from .frontend.features import pre_emphasis
from .frontend.features import trfbank
from .frontend.features import mel_filter_bank
from .frontend.features import mfcc
from .frontend.features import pca_dct
from .frontend.features import shifted_delta_cepstral

from .iv_scoring import cosine_scoring
from .iv_scoring import mahalanobis_scoring
from .iv_scoring import two_covariance_scoring
from .iv_scoring import PLDA_scoring

from .gmm_scoring import gmm_scoring

from .jfa_scoring import jfa_scoring

from .sidekit_io import write_norm_hdf5
from .sidekit_io import write_matrix_hdf5


from .sv_utils import clean_stat_server

libsvm_loaded = False
if SIDEKIT_CONFIG["libsvm"]:
    try:
        dirname = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'libsvm')
        if sys.platform == 'win32':
            libsvm = CDLL(os.path.join(dirname, r'libsvm.dll'))
            libsvm_loaded = True
        else:
            libsvm = CDLL(os.path.join(dirname, 'libsvm.so.2'))
            libsvm_loaded = True
    except:
        # For unix the prefix 'lib' is not considered.
        if find_library('svm'):
            libsvm = CDLL(find_library('svm'))
            libsvm_loaded = True
        elif find_library('libsvm'):
            libsvm = CDLL(find_library('libsvm'))
            libsvm_loaded = True
        else:
            libsvm_loaded = False
            logging.warning('WARNNG: libsvm is not installed, please refer to the' +
                            ' documentation if you intend to use SVM classifiers')

if libsvm_loaded:
    from sidekit.libsvm import *
    from sidekit.svm_scoring import *
    from sidekit.svm_training import *


CUDA = False
if SIDEKIT_CONFIG["cuda"]:
    CUDA = True


if CUDA:
    from .nnet import FForwardNetwork
    from .nnet import kaldi_to_hdf5
    from .nnet import XvectorMultiDataset
    from .nnet import XvectorDataset
    from .nnet import StatDataset
    from .nnet import Xtractor
    from .nnet import xtrain
    from .nnet import xtrain_asynchronous
    from .nnet import extract_idmap
    from .nnet import extract_parallel
    from .nnet import extract_embeddings
    from .nnet import ResBlock
    from .nnet import ResNet18

else:
    print("Don't import Torch")

if SIDEKIT_CONFIG["mpi"]:
    found_mpi4py = importlib.find_loader('mpi4py') is not None
    if found_mpi4py:
        from .sidekit_mpi import EM_split
        from .sidekit_mpi import total_variability
        from .sidekit_mpi import extract_ivector
        print("Import MPI")
        

__author__ = "Anthony Larcher and Sylvain Meignier"
__copyright__ = "Copyright 2014-20120 Anthony Larcher and Sylvain Meignier"
__license__ = "LGPL"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'
__version__="1.3.6.6"




# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#    
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as 
# published by the Free Software Foundation, either version 3 of the License, 
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2019 Anthony Larcher

:mod:`sidekit_wrappers` provides wrappers for different purposes.
The aim when using wrappers is to simplify the development of new function
in an efficient manner
"""
import os
import numpy
import copy
import logging
from sidekit import PARALLEL_MODULE

__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2014-2019 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


def coroutine(func):
    """
    Decorator that allows to forget about the first call of a coroutine .next()
    method or .send(None)
    This call is done inside the decorator
    :param func: the coroutine to decorate
    """
    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        next(cr)
        return cr
    return start


def deprecated(func):
    """

    :param func:
    :return:
    """
    count = [0]

    def wrapper(*args, **kwargs):
        count[0] += 1
        if count[0] == 1:
            logging.warning(func.__name__ + ' is deprecated')
        return func(*args, **kwargs)
    return wrapper


def check_path_existance(func):
    """ Decorator for a function wich prototype is:
    
        func(features, outputFileName)
        
        This decorator gets the path included in 'outputFileName' if any 
        and check if this path exists; if not the path is created.
        :param func: function to decorate
    """
    def wrapper(*args, **kwargs):
        dir_name = os.path.dirname(args[1])  # get the path
        # Create the directory if it dosn't exist
        if not os.path.exists(dir_name) and (dir_name is not ''):
            os.makedirs(dir_name)            
        # Do the job
        func(*args, **kwargs)
    return wrapper


def process_parallel_lists(func):
    """
    Decorator that is used to parallelize process.
    This decorator takes a function with any list of arguments including 
    "num_thread" and parallelize the process by creating "num_thread" number of
    parallel process or threads.
    
    The choice of process or threas depends on the value of the global variable
    "PARALLEL_MODULE" that is defined in  ./sidekit/__init__.py
      
    Parallelization is done as follow:
        - all arguments have to be given to the decorator with their names
          any other case might limit the parallelization.
        - the function that is decorated is called by "num_thread" concurrent
          process (or threads) with the list of arguments that is given 
          to the decorator except special arguments (see below)

    Special arguments:
        Special arguments are the one that lead to parallelization.
        There are 3 types of special arguments which name end with a special 
        suffix:
        
        - arguments which names are "*_list" or "*_indices" are lists 
          or numpy arrays that will be split (equally or almost) and
          each sub-list will be passed as an argument for a process/thread
        
        - arguments which names are "_acc" are duplicated and each thread is
          given a copy of this accumulator. At the end of the function, all
          accumulators will be summed to return a unique accumulatore; thus
          any object passed as a "*_acc" argument has to implement 
          a "+" operator

        - arguments which names are "*_server" are duplicated using a deepcopy
          of the original argument. This is mostly used to pass servers such
          as FeaturesServer as arguments
    :param func: function to decorate
    
    """
    def wrapper(*args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        
        if len(args) > 1:
            print("Warning, some arguments are not named, computation might not be parallelized")
        
        num_thread = 1
        if "num_thread" in kwargs.keys():
            num_thread = kwargs["num_thread"]
        
        # On créé un dictionnaire de paramètres kwargs pour chaque thread
        if PARALLEL_MODULE in ['threading', 'multiprocessing'] and num_thread > 1:

            # If arguments end with _list or _indices,
            # set number of Threads to the minimum length of the lists and raise a warning
            list_length = numpy.inf
            for k, v in kwargs.items():
                # If v is a list or a numpy.array
                if k.endswith("_list") or k.endswith("_indices"):
                    list_length = min(list_length, len(list(v)))
            num_thread = min(num_thread, list_length)

            # Create a list of dictionaries, one per thread, and initialize
            # them with the keys
            parallel_kwargs = []
            for ii in range(num_thread):
                parallel_kwargs.append(dict(zip(kwargs.keys(), 
                                            [None]*len(kwargs.keys()))))
 
            for k, v in kwargs.items():
                
                # If v is a list or a numpy.array
                if k.endswith("_list") or k.endswith("_indices"):
                    sub_lists = numpy.array_split(v, num_thread)
                    for ii in range(num_thread):
                        parallel_kwargs[ii][k] = sub_lists[ii]  # the ii-th sub_list is used for the thread ii

                elif k == "num_thread":
                    for ii in range(num_thread):
                        parallel_kwargs[ii][k] = 1
 
                # If v is an accumulator (meaning k ends with "_acc")
                # v is duplicated for each thread
                elif k.endswith("_acc"):
                    for ii in range(num_thread):
                        parallel_kwargs[ii][k] = v

                # Duplicate servers for each thread
                elif k.endswith("_server") or k.endswith("_extractor"):
                    for ii in range(num_thread):
                        parallel_kwargs[ii][k] = copy.deepcopy(v)
                        
                # All other parameters are just given to each thread
                else:
                    for ii in range(num_thread):
                        parallel_kwargs[ii][k] = v
            
            if PARALLEL_MODULE is 'multiprocessing':
                import multiprocessing
                jobs = []
                multiprocessing.freeze_support()
                for idx in range(num_thread):
                    p = multiprocessing.Process(target=func, args=args, kwargs=parallel_kwargs[idx])
                    jobs.append(p)
                    p.start()
                for p in jobs:
                    p.join()
            
            elif PARALLEL_MODULE is 'threading':
                import threading
                jobs = []
                for idx in range(num_thread):
                    p = threading.Thread(target=func, args=args, kwargs=parallel_kwargs[idx])
                    jobs.append(p)
                    p.start()
                for p in jobs:
                    p.join()
        
            elif PARALLEL_MODULE is 'MPI':
                # TODO
                print("ParallelProcess using MPI is not implemented yet")
                pass
           
            # Sum accumulators if any
            for k, v in kwargs.items():
                if k.endswith("_acc"):
                    for ii in range(num_thread):
                        if isinstance(kwargs[k], list):
                            kwargs[k][0] += parallel_kwargs[ii][k][0]
                        else:
                            kwargs[k] += parallel_kwargs[ii][k]

        else:
            logging.debug("No Parallel processing with this module")
            func(*args, **kwargs)
        
    return wrapper


def accepts(*types, **kw):
    """Function decorator. Checks decorated function's arguments are
    of the expected types.
    
    Sources: https://wiki.python.org/moin/PythonDecoratorLibrary#Type_Enforcement_.28accepts.2Freturns.29
    
    Parameters:
        types -- The expected types of the inputs to the decorated function.
            Must specify type for each parameter.
        kw    -- Optional specification of 'debug' level (this is the only valid
            keyword argument, no other should be given).
            debug = ( 0 | 1 | 2 )
            
    """
    if not kw:
        # default level: MEDIUM
        debug = 1
    else:
        debug = kw['debug']
    try:
        def decorator(f):
            def newf(*args):
                if debug is 0:
                    return f(*args)
                assert len(args) == len(types)
                argtypes = tuple([a.__class__.__name__ for a in args])
                if argtypes != types:
                    print("argtypes = {} and types = {}".format(argtypes, types))
                    msg = info(f.__name__, types, argtypes, 0)
                    if debug is 1:
                        print('TypeWarning: ', msg)
                    elif debug is 2:
                        raise TypeError(msg)
                return f(*args)
            newf.__name__ = f.__name__
            return newf
        return decorator
    except KeyError as key:
        raise KeyError(key + "is not a valid keyword argument")
    except TypeError(msg):
        raise TypeError(msg)


def info(fname, expected, actual, flag):
    """Convenience function returns nicely formatted error/warning msg.
    :param fname: function to decorate
    :param expected: expected format of the function
    :param actual: actual format of the function to check
    :param flag: flag
    """
    _format = lambda types: ', '.join([str(t).split("'")[0] for t in types])
    expected, actual = _format(expected), _format(actual)
    msg = "'{}' method ".format(fname)\
          + ("accepts", "returns")[flag] + " ({}), but ".format(expected)\
          + ("was given", "result is")[flag] + " ({})".format(actual)
    return msg



# -*- coding: utf-8 -*-

# This package is a translation of a part of the BOSARIS toolkit.
# The authors thank Niko Brummer and Agnitio for allowing them to
# translate this code and provide the community with efficient structures
# and tools.
#
# The BOSARIS Toolkit is a collection of functions and classes in Matlab
# that can be used to calibrate, fuse and plot scores from speaker recognition
# (or other fields in which scores are used to test the hypothesis that two
# samples are from the same source) trials involving a model and a test segment.
# The toolkit was written at the BOSARIS2010 workshop which took place at the
# University of Technology in Brno, Czech Republic from 5 July to 6 August 2010.
# See the User Guide (available on the toolkit website)1 for a discussion of the
# theory behind the toolkit and descriptions of some of the algorithms used.
#
# The BOSARIS toolkit in MATLAB can be downloaded from `the website
# <https://sites.google.com/site/bosaristoolkit/>`_.

"""
This is the 'plotwindow' module
"""
import numpy


__author__ = "Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'
__credits__ = ["Niko Brummer", "Edward de Villiers"]


class PlotWindow:
    """A class that is used to define the parameters of a plotting window.

    :attr __pfa_limits__: ndarray of two values that determine the limits of the pfa axis.
            Default is [0.0005, 0.5]
    :attr __pmiss_limits__: ndarray of two values that determine the limits of the pmiss axis.
                Default is [0.0005, 0.5]
    :attr __xticks__: coordonates of the ticks on the horizontal axis
    :attr __xticklabels__: labels of the ticks on the horizontal axis in a ndarray of strings
    :attr __yticks__: coordonates of the ticks on the vertical axis
    :attr __yticklabels__: labels of the ticks on the vertical axis in a ndarray of strings
    """

    def __init__(self, input_type=''):
        """Initialize PlotWindow object to one of the pre-defined ploting type.
        - 'new'
        - 'old'
        - 'big'
        - 'sre10'

        :param input_type: the type of DET plot to display. Default is 'old'
        """

        if input_type == '':
            self.__pfa_limits__ = numpy.array([5e-4, 5e-1])
            self.__pmiss_limits__ = numpy.array([5e-4, 5e-1])
            self.__xticks__ = numpy.array([0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3, 0.4])
            self.__xticklabels__ = numpy.array(['0.1', '0.2', '0.5', ' 1 ', ' 2 ', ' 5 ', '10 ', '20 ', '30 ', '40 '])
            self.__yticks__ = numpy.array([0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3, 0.4])
            self.__yticklabels__ = numpy.array(['0.1', '0.2', '0.5', ' 1 ', ' 2 ', ' 5 ', '10 ', '20 ', '30 ', '40 '])
        elif input_type == 'new':
            self.axis_new()
        elif input_type == 'old':
            self.axis_old()
        elif input_type == 'big':
            self.axis_big()
        elif input_type == 'sre10':
            self.axis_sre10()
        else:
            raise Exception('Error, wrong type of PlotWindow')

    def make_plot_window_from_values(self, pfa_limits, pmiss_limits, xticks, xticklabels, yticks, yticklabels):
        """Initialize PlotWindow from provided values

        :param pfa_limits: ndarray of two values that determine the limits of the pfa axis.
        :param pmiss_limits: ndarray of two values that determine the limits of the pmiss axis.
        :param xticks: coordonates of the ticks on the horizontal axis.
        :param xticklabels: labels of the ticks on the horizontal axis in a ndarray of strings.
        :param yticks: coordonates of the ticks on the vertical axis.
        :param yticklabels: labels of the ticks on the vertical axis in a ndarray of strings.
        """
        self.__pfa_limits__ = pfa_limits
        self.__pmiss_limits__ = pmiss_limits
        self.__xticks__ = xticks
        self.__xticklabels__ = xticklabels
        self.__yticks__ = yticks
        self.__yticklabels__ = yticklabels

    def axis_new(self):
        """Set axis value to new ones

        - pfa ranges from 0.000005 to 0.005
        - pmiss ranges from 0.01 to 0.99
        """
        self.__pfa_limits__ = numpy.array([5e-6, 5e-3])
        self.__pmiss_limits__ = numpy.array([1e-2, 0.99])
        self.__xticks__ = numpy.array([1e-5, 2e-5, 5e-5, 1e-4, 2e-4, 5e-4, 1e-3, 2e-3])
        self.__xticklabels__ = numpy.array(['1e-3', '2e-3', '5e-3', '0.01', '0.02', '0.05', '0.1 ', '0.2 '])
        self.__yticks__ = numpy.array([0.02, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.98])
        self.__yticklabels__ = numpy.array([' 2 ', ' 5 ', '10 ', '20 ', '30 ',
                                            '40 ', '50 ', '60 ', '70 ', '80 ', '90 ', '95 ', '98 '])

    def axis_old(self):
        """Set axis value to old ones (NIST-SRE08 style)

        - pfa ranges from 0.0005 to 0.5
        - pmiss ranges from 0.0005 to 0.5
        """
        self.__pfa_limits__ = numpy.array([5e-4, 5e-1])
        self.__pmiss_limits__ = numpy.array([5e-4, 5e-1])
        self.__xticks__ = numpy.array([0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3, 0.4])
        self.__xticklabels__ = numpy.array(['0.1', '0.2', '0.5', ' 1 ', ' 2 ', ' 5 ', '10 ', '20 ', '30 ', '40 '])
        self.__yticks__ = numpy.array([0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3, 0.4])
        self.__yticklabels__ = numpy.array(['0.1', '0.2', '0.5', ' 1 ', ' 2 ', ' 5 ', '10 ', '20 ', '30 ', '40 '])

    def axis_big(self):
        """Set axis value to big ones

        - pfa ranges  from 0.000005 to 0.99
        - pmiss ranges from 0.000005 to0.99
        """
        self.__pfa_limits__ = numpy.array([5e-6, 0.99])
        self.__pmiss_limits__ = numpy.array([5e-6, 0.99])
        self.__yticks__ = numpy.array([5e-6, 5e-5, 5e-4, 0.5e-2, 2.5e-2, 10e-2,
                                       25e-2, 50e-2, 72e-2, 88e-2, 96e-2, 99e-2])
        self.__yticklabels__ = numpy.array(['5e-4', '5e-3', '0.05', '0.5 ',
                                            '2.5 ', ' 10 ', ' 25 ', ' 50 ', ' 72 ', ' 88 ', ' 96 ', ' 99 '])
        self.__xticks__ = numpy.array([5e-5, 5e-4, 0.5e-2, 2.5e-2, 10e-2, 25e-2, 50e-2, 72e-2, 88e-2, 96e-2, 99e-2])
        self.__xticklabels__ = numpy.array(['5e-3', '0.05', '0.5 ', '2.5 ',
                                            ' 10 ', ' 25 ', ' 50 ', ' 72 ', ' 88 ', ' 96 ', ' 99 '])

    def axis_sre10(self):
        """Set axis value to NIST-SRE10 style

        - pfa ranges from 0.000003 to 0.5
        - pmiss ranges from 0.0003 to 0.9
        """
        self.__pfa_limits__ = numpy.array([3e-6, 5e-1])
        self.__pmiss_limits__ = numpy.array([3e-4, 9e-1])
        self.__xticks__ = numpy.array([1e-5, 1e-4, 1e-3, 2e-3, 5e-3, 1e-2, 2e-2, 5e-2, 1e-1, 2e-1, 4e-1])
        self.__xticklabels__ = numpy.array(['0.001', ' 0.01', '  0.1', '  0.2',
                                            '  0.5', '    1', '    2', '    5', '   10', '   20', '   40'])
        self.__yticks__ = numpy.array([1e-3, 2e-3, 5e-3, 1e-2, 2e-2, 5e-2, 1e-1, 2e-1, 4e-1, 8e-1])
        self.__yticklabels__ = numpy.array(['0.1', '0.2', '0.5', ' 1 ', ' 2 ', ' 5 ', ' 10', ' 20', ' 40', ' 80'])



# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#    
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as 
# published by the Free Software Foundation, either version 3 of the License, 
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2019 Anthony Larcher

:mod:`sv_utils` provides utilities to facilitate the work with SIDEKIT.
"""
import ctypes
import copy
import gzip
import multiprocessing
import numpy
import os
import pickle
import re
import scipy
import sys
if sys.version_info.major > 2:
    from functools import reduce

__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2014-2019 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


def save_svm(svm_file_name, w, b):
    """Save SVM weights and bias in PICKLE format
    
    :param svm_file_name: name of the file to write
    :param w: weight coefficients of the SVM to store
    :param b: biais of the SVM to store
    """
    if not os.path.exists(os.path.dirname(svm_file_name)):
            os.makedirs(os.path.dirname(svm_file_name))
    with gzip.open(svm_file_name, "wb") as f:
            pickle.dump((w, b), f)


def read_svm(svm_file_name):
    """Read SVM model in PICKLE format
    
    :param svm_file_name: name of the file to read from
    
    :return: a tupple of weight and biais
    """
    with gzip.open(svm_file_name, "rb") as f:
        (w, b) = pickle.load(f)
    return numpy.squeeze(w), b


def check_file_list(input_file_list, file_name_structure):
    """Check the existence of a list of files in a specific directory
    Return a new list with the existing segments and a list of indices 
    of those files in the original list. Return outputFileList and 
    idx such that inputFileList[idx] = outputFileList
    
    :param input_file_list: list of file names
    :param file_name_structure: structure of the filename to search for
    
    :return: a list of existing files and the indices 
        of the existing files in the input list
    """
    exist_files = numpy.array([os.path.isfile(file_name_structure.format(f)) for f in input_file_list])
    output_file_list = input_file_list[exist_files]
    idx = numpy.argwhere(numpy.in1d(input_file_list, output_file_list))
    return output_file_list, idx.transpose()[0]


def initialize_iv_extraction_weight(ubm, T):
    """
    Estimate matrices W and T for approximation of the i-vectors
    For more information, refers to [Glembeck09]_

    :param ubm: Mixture object, Universal Background Model
    :param T: Raw TotalVariability matrix as a ndarray
    
    :return:    
      W: fix matrix pre-computed using the weights from the UBM and the 
          total variability matrix
      Tnorm: total variability matrix pre-normalized using the co-variance 
          of the UBM
    """
    # Normalize the total variability matrix by using UBM co-variance
    
    sqrt_invcov = numpy.sqrt(ubm.get_invcov_super_vector()[:, numpy.newaxis])
    Tnorm = T * sqrt_invcov
    
    # Split the Total Variability matrix into sub-matrices per distribution
    Tnorm_c = numpy.array_split(Tnorm, ubm.distrib_nb())
    
    # Compute fixed matrix W
    W = numpy.zeros((T.shape[1], T.shape[1]))
    for c in range(ubm.distrib_nb()):
        W = W + ubm.w[c] * numpy.dot(Tnorm_c[c].transpose(), Tnorm_c[c])

    return W, Tnorm


def initialize_iv_extraction_eigen_decomposition(ubm, T):
    """Estimate matrices Q, D_bar_c and Tnorm, for approximation 
    of the i-vectors.
    For more information, refers to [Glembeck09]_
    
    :param ubm: Mixture object, Universal Background Model
    :param T: Raw TotalVariability matrix
    
    :return:
      Q: Q matrix as described in [Glembeck11]
      D_bar_c: matrices as described in [Glembeck11]
      Tnorm: total variability matrix pre-normalized using the co-variance of the UBM
    """
    # Normalize the total variability matrix by using UBM co-variance
    sqrt_invcov = numpy.sqrt(ubm.get_invcov_super_vector()[:, numpy.newaxis])
    Tnorm = T * sqrt_invcov
    
    # Split the Total Variability matrix into sub-matrices per distribution
    Tnorm_c = numpy.array_split(Tnorm, ubm.distrib_nb())
    
    # Compute fixed matrix Q
    W = numpy.zeros((T.shape[1], T.shape[1]))
    for c in range(ubm.distrib_nb()):
        W = W + ubm.w[c] * numpy.dot(Tnorm_c[c].transpose(), Tnorm_c[c])
    
    eigen_values, Q = scipy.linalg.eig(W)
    
    # Compute D_bar_c matrix which is the diagonal approximation of Tc' * Tc
    D_bar_c = numpy.zeros((ubm.distrib_nb(), T.shape[1]))
    for c in range(ubm.distrib_nb()):
        D_bar_c[c, :] = numpy.diag(reduce(numpy.dot, [Q.transpose(), Tnorm_c[c].transpose(), Tnorm_c[c], Q]))
    return Q, D_bar_c, Tnorm


def initialize_iv_extraction_fse(ubm, T):
    """Estimate matrices for approximation of the i-vectors.
    For more information, refers to [Cumani13]_
    
    :param ubm: Mixture object, Universal Background Model
    :param T: Raw TotalVariability matrix
    
    :return:
      Q: Q matrix as described in [Glembeck11]
      D_bar_c: matrices as described in [Glembeck11]
      Tnorm: total variability matrix pre-normalized using the co-variance of the UBM
    """
    # % Initialize the process
    # %init = 1;
    #
    #
    #   Extract i-vectors by using different methods
    #
    # %rank_T      = 10;
    # %featureSize = 50;
    # %distribNb   = 32;
    # %dictSize    = 5;
    # %dictSizePerDis=rank_T;  % a modifier par la suite
    #
    # %ubm_file    = 'gmm/world32.gmm';
    # %t_file      = 'mat/TV_32.matx';
    #
    #
    #
    #   Load data
    #
    #
    # % Load UBM for weight parameters that are used in the optimization
    # % function
    # %UBM = ALize_LoadModel(ubm_file);
    #
    # % Load meand from Minimum Divergence re-estimation
    # %sv_mindiv = ALize_LoadVect(minDiv_file)';
    #
    # % Load T matrix
    # %T = ALize_LoadMatrix(t_file)';
    #
    #
    # function [O,PI,Q] = factorized_subspace_estimation(UBM,T,dictSize,outIterNb,inIterNb)
    #
    #    rank_T      = size(T,2);
    #    distribNb   = size(UBM.W,2);
    #    featureSize = size(UBM.mu,1);
    #
    #    % Normalize matrix T
    #    sv_invcov = reshape(UBM.invcov,[],1);
    #    T_norm = T .* repmat(sqrt(sv_invcov),1,rank_T);
    #
    #     Tc{1,distribNb} = [];
    #     for ii=0:distribNb-1
    #         % Split the matrix in sub-matrices
    #         Tc{ii+1} = T_norm((ii*featureSize)+1:(ii+1)*featureSize,:);
    #     end
    #
    #     % Initialize O and Qc by Singular Value Decomposition
    #     init_FSE.Qc{distribNb} 	= [];
    #     init_FSE.O{distribNb}   = [];
    #     PI{distribNb}           = [];
    #
    #     for cc=1:distribNb
    #         init_FSE.Qc{cc} 	= zeros(featureSize,featureSize);
    #         init_FSE.O{cc}   = zeros(featureSize,featureSize);
    #         PI{cc}  = sparse(zeros(featureSize,dictSize*distribNb));      % a remplacer par une matrice sparse
    #     end
    #
    #     % For each distribution
    #     for cc=1:distribNb
    #         fprintf('Initilize matrice for distribution %d / %d\n',cc,distribNb);
    #         % Initialized O with Singular vectors from SVD
    #         [init_FSE.O{cc},~,V] = svd(UBM.W(1,cc)*Tc{cc});
    #         init_FSE.Qc{cc} = V'; 
    #     end
    #
    #
    #    % Concatenate Qc to create the matrix Q: A MODIFIER POUR DISSOCIER
    #    % dictSize DU NOMBRE DE DISTRIBUTIONS
    #    Q = [];
    #    for cc=1:distribNb
    #        Q = [Q;init_FSE.Qc{cc}(1:dictSize,:)];
    #    end
    #    O = init_FSE.O;
    #    clear 'init_FSE'
    #
    #
    #    % OUTER iteration process : update Q iteratively
    #    for it = 1:outIterNb
    #
    #        fprintf('Start iteration %d / %d for Q re-estimation\n',it,5);
    #
    #        % INNER iteration process: update PI and O iteratively
    #        for pioIT = 1:inIterNb
    #            fprintf('   Start iteration %d / %d for PI and O re-estimation\n',pioIT,10);
    #
    #            % Update PI
    #            %Compute diagonal terms of QQ'
    # %            diagQ = diag(Q*Q');
    #
    #            for cc=1:distribNb
    #
    #                % Compute optimal k and optimal v 
    #                % for each line f of PI{cc}
    #                A = O{cc}'*Tc{cc}*Q';
    #                f = 1;
    #                while (f < size(A,1)+1)
    #
    #                    if(f == 1)
    #                        A = O{cc}'*Tc{cc}*Q';       % equation (26)
    #                        PI{cc} = sparse(zeros(featureSize,dictSize*distribNb));
    #                    end
    #
    #                    % Find the optimal index k
    #                    [~,k] = max(A(f,:).^2);
    #                    k_opt = k;
    #
    #                    % Find the optimal value v
    #                    v_opt = A(f,k_opt);
    #
    #                    % Update the line of PI{cc} with the value v_opt in the
    #                    % k_opt-th column
    #                    PI{cc}(f,k_opt)     = v_opt;
    #
    #                    % if the column already has a non-zero element,
    #                    % update O and PI
    #                    I = find(PI{cc}(:,k_opt)~=0);
    #                    if size(I,1)>1
    #                        % get indices of the two lines of PI{cc} which
    #                        % have a non-zero element on the same column
    #                        a = I(1);
    #                        b = I(2);
    #
    #                        % Replace column O{cc}(:,a) and O{cc}(:,b)
    #                        Oa = (PI{cc}(a,k_opt)*O{cc}(:,a)+PI{cc}(b,k_opt)*O{cc}(:,b))/
    # (sqrt(PI{cc}(a,k_opt)^2+PI{cc}(b,k_opt)^2));
    #                        Ob = (PI{cc}(a,k_opt)*O{cc}(:,b)-PI{cc}(b,k_opt)*O{cc}(:,a))/
    # (sqrt(PI{cc}(a,k_opt)^2+PI{cc}(b,k_opt)^2));
    #                        O{cc}(:,a) = Oa;
    #                        O{cc}(:,b) = Ob;
    #
    #                        PI{cc}(a,k_opt) = sqrt(PI{cc}(a,k_opt)^2+PI{cc}(b,k_opt)^2);
    #                        PI{cc}(b,k_opt) = 0;
    #
    #                        f = 0;
    #                    end
    #                    f =f +1;
    #                end
    #            end
    #
    #            obj = computeObjFunc(UBM.W,Tc,O,PI,Q);
    #            fprintf('Objective Function after estimation of PI = %2.10f\n',obj);                
    #
    #            % Update O
    #            for cc=1:distribNb
    #
    #                % Compute 
    #                Z = PI{cc}*Q*Tc{cc}';
    #
    #                % Compute Singular value decomposition of Z
    #                [Uz,~,Vz] = svd(Z);
    #
    #                % Compute the new O{cc}
    #                O{cc} = Vz*Uz';
    #            end
    #            obj = computeObjFunc(UBM.W,Tc,O,PI,Q);
    #            fprintf('Objective Function after estimation of O = %2.10f\n',obj);
    #        end % END OF INNER ITERATION PROCESS
    #
    #
    #        % Update Q
    #        D = sparse(zeros(size(PI{cc},2),size(PI{cc},2)));
    #        E = zeros(size(PI{cc},2),rank_T);
    #        for cc=1:distribNb
    #            % Accumulate D
    #            D = D + UBM.W(1,cc) * PI{cc}'*PI{cc};
    #
    #            % Accumulate the second term
    #            E = E + UBM.W(1,cc) * PI{cc}'*O{cc}'*Tc{cc};
    #        end
    #        Q = D\E;
    #
    #        % Normalize rows of Q and update PI accordingly
    #
    #        % Compute norm of each row
    #        c1 = bsxfun(@times,Q,Q);
    #        c2 = sum(c1,2);
    #        c3 = sqrt(c2);
    #        Q = bsxfun(@rdivide,Q,c3);
    #
    #        % Update PI accordingly
    #        PI = cellfun(@(x)x*sparse(diag(c3)), PI, 'uni',false);
    #
    #        obj = computeObjFunc(UBM.W,Tc,O,PI,Q);
    #        fprintf('Objective Function after re-estimation of Q = %2.10f\n',obj);
    #    end
    # end
    pass


def clean_stat_server(statserver):
    """

    :param statserver:
    :return:
    """
    zero_idx = ~(statserver.stat0.sum(axis=1) == 0.)
    statserver.modelset = statserver.modelset[zero_idx]
    statserver.segset = statserver.segset[zero_idx]
    statserver.start = statserver.start[zero_idx]
    statserver.stop = statserver.stop[zero_idx]
    statserver.stat0 = statserver.stat0[zero_idx, :]
    statserver.stat1 = statserver.stat1[zero_idx, :]
    assert statserver.validate(), "Error after cleaning StatServer"
    print("Removed {} empty sessions in StatServer".format((~zero_idx).sum()))


def parse_mask(mask):
    """

    :param mask:
    :return:
    """
    if not set(re.sub("\s", "", mask)[1:-1]).issubset(set("0123456789-,")):
        raise Exception("Wrong mask format")
    tmp = [k.split('-') for k in re.sub(r"[\s]", '', mask)[1:-1].split(',')]
    indices = []
    for seg in tmp:
        if len(seg) == 1:
            seg += seg
        if len(seg) == 2:
            indices += list(range(int(seg[0]), int(seg[1])+1))
        else:
            raise Exception("Wrong mask format")
    return indices


def segment_mean_std_hdf5(input_segment, in_context=False):
    """
    Compute the sum and square sum of all features for a list of segments.
    Input files are in HDF5 format

    :param input_segment: list of segments to read from, each element of the list is a tuple of 5 values,
        the filename, the index of thefirst frame, index of the last frame, the number of frames for the
        left context and the number of frames for the right context
    :param in_context:
    :return: a tuple of three values, the number of frames, the sum of frames and the sum of squares
    """
    features_server, show, start, stop, in_context = input_segment

    if start is None or stop is None or not in_context:
        feat, _ = features_server.load(show,
                                       start=start,
                                       stop=stop)

    else:
        # Load the segment of frames plus left and right context
        feat, _ = features_server.load(show,
                                       start=start-features_server.context[0],
                                       stop=stop+features_server.context[1])
        # Get features in context
        feat, _ = features_server.get_context(feat=feat,
                                              label=None,
                                              start=features_server.context[0],
                                              stop=feat.shape[0]-features_server.context[1])

    return feat.shape[0], feat.sum(axis=0), numpy.sum(feat**2, axis=0)


def mean_std_many(features_server, seg_list, in_context=False, num_thread=1):
    """
    Compute the mean and standard deviation from a list of segments.

    :param features_server:
    :param seg_list: list of file names with start and stop indices
    :param in_context:
    :param num_thread:
    :return: a tuple of three values, the number of frames, the mean and the variance
    """
    if isinstance(seg_list[0], tuple):
        inputs = [(copy.deepcopy(features_server), seg[0], seg[1], seg[2], in_context) for seg in seg_list]
    elif isinstance(seg_list[0], str):
        inputs = [(copy.deepcopy(features_server), seg, None, None, in_context) for seg in seg_list]

    pool = multiprocessing.Pool(processes=num_thread)
    res = pool.map(segment_mean_std_hdf5, inputs)
    pool.terminate()
    total_N = 0
    total_F = 0
    total_S = 0
    for N, F, S in res:
        total_N += N
        total_F += F
        total_S += S
    return total_N, total_F / total_N, total_S / total_N


def serialize(M):
    M_shape = M.shape
    ct = ctypes.c_double
    if M.dtype == numpy.float32:
        ct = ctypes.c_float
    tmp_M = multiprocessing.Array(ct, M.size)
    M = numpy.ctypeslib.as_array(tmp_M.get_obj())
    return M.reshape(M_shape)



# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2020 Anthony Larcher
"""

import h5py
import logging
import sys
import numpy
import pickle
import torch
import torch.optim as optim
import torch.multiprocessing as mp
from torchvision import transforms
from collections import OrderedDict
from .xsets import XvectorMultiDataset, XvectorDataset, StatDataset, VoxDataset
from .xsets import FrequencyMask, CMVN, TemporalMask
from ..bosaris import IdMap
from ..statserver import StatServer
from torch.utils.data import DataLoader

class ResBlock(torch.nn.Module):
    """

    """
    def __init__(self, in_channels, out_channels, is_first=False):
        """

        :param filter_size:
        :param channel_nb:
        :param is_first: boolean, True if this block ios the first of the model, if not, apply a BatchNorm layer first
        """
        super(ResBlock, self).__init__()
        self.is_first = is_first
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.expansion = self.out_channels // self.in_channels

        self.resample = None
        if not self.in_channels == self.out_channels:
            self.resample = torch.nn.Sequential(
                torch.nn.Conv2d(in_channels=self.in_channels,
                                out_channels=self.out_channels,
                                kernel_size=1),
                torch.nn.BatchNorm2d(self.in_channels * self.expansion),
            )

        if not self.is_first:
            self.batch_norm1 = torch.nn.BatchNorm2d(num_features=self.in_channels)

        self.activation = torch.nn.LeakyReLU()

        self.conv1 = torch.nn.Conv2d(in_channels=self.in_channels,
                                     out_channels=self.out_channels,
                                     kernel_size=(3,3),
                                     padding=1,
                                     padding_mode='zeros',
                                     dilation=1)
        self.conv2= torch.nn.Conv2d(in_channels=self.out_channels,
                                    out_channels=self.out_channels,
                                    kernel_size=(3,3),
                                    padding=1,
                                    padding_mode='zeros',
                                    dilation=1)

        self.batch_norm2 = torch.nn.BatchNorm2d(num_features=self.out_channels)

    def forward(self, x):
        """

        :param x:
        :return:
        """
        identity = x

        if not self.is_first:
            out = self.activation(self.batch_norm1(x))
        else:
            out = x

        out = self.conv1(out)
        out = self.batch_norm2(out)
        out = self.activation(out)

        out = self.conv2(out)
        out = self.batch_norm2(out)

        if not self.expansion == 1:
            identity = self.resample(identity)
        out += identity

        out =  self.activation(out)
        return out


class ResNet18(torch.nn.Module):
    """

    """
    def __init__(self, spk_number,
                 entry_conv_kernel_size=(7,7),
                 entry_conv_out_channels=64,
                 megablock_out_channels=(64, 128, 256, 512),
                 megablock_size=(2, 2, 2, 2),
                 block_type = ResBlock
                 ):
        """

        :param spk_number:
        :param entry_conv_kernel_size:
        :param entry_conv_out_channels:
        :param megablock_out_channels:
        :param megablock_size:
        :param block_type:
        """
        super(ResNet18, self).__init__()

        self.spk_number = spk_number
        self.activation = torch.nn.LeakyReLU()

        # First convolution layer for input
        self.entry_conv = torch.nn.Conv2d(in_channels=1,
                                          out_channels=entry_conv_out_channels,
                                          kernel_size=entry_conv_kernel_size,
                                          padding=3,
                                          stride=1)
        self.entry_batch_norm = torch.nn.BatchNorm2d(entry_conv_out_channels)
        self.top_channel_number = entry_conv_out_channels

        # Add ResBlocks
        self.mega_blocks = []
        for mb_size, mb_out in zip(megablock_size, megablock_out_channels):
            self.mega_blocks.append(self._add_megablock(block_type, mb_size, mb_out))
            self.top_channel_number = mb_out

        # Top layers for classification and embeddings extraction
        self.top_lin1 = torch.nn.Linear(megablock_out_channels[-1] * 2 * 40, 512)  # a modifier pour voir la taille
        self.top_batch_norm1 = torch.nn.BatchNorm1d(512)
        self.top_lin2 = torch.nn.Linear(512, spk_number)

    def forward(self, x):
        """

        :param x:
        :return:
        """
        print("entree: {}".format(x.shape))
        x = self.entry_conv(x)
        x = self.entry_batch_norm(x)
        x = self.activation(x)

        print("Avant resblocks: {}".format(x.shape))
        for layer in self.mega_blocks:
            x = layer(x)

        print("Avant pooling: {}".format(x.shape))
        # Pooling done as for x-vectors
        mean = torch.mean(x, dim=2)
        mean = torch.flatten(mean, 1)
        std = torch.std(x, dim=2)
        std = torch.flatten(std, 1)
        x = torch.cat([mean, std], dim=1)

        # Classification layers
        x = self.top_lin1(x)
        x = self.top_batch_norm1(x)
        x = self.activation(x)
        x = self.top_lin2(x)

        return x

    def _add_megablock(self, block_type, block_nb, out_channels, is_first=False):

        rblocks = [block_type(self.top_channel_number, out_channels, is_first=is_first),]
        for _ in range(1, block_nb):
            rblocks.append(block_type(out_channels, out_channels, is_first=False))
        return torch.nn.Sequential(*rblocks)


def restrain(args):
    """
    Initialize and train an ResNet for Speaker Recognition

    :param args:
    :return:
    """
    # Initialize a first model and save to disk
    model = ResNet18(args.class_number,
                     entry_conv_kernel_size=(7,7),
                     entry_conv_out_channels=64,
                     megablock_out_channels=(64, 128, 128, 128),
                     megablock_size=(2, 2, 2, 2),
                     block_type = ResBlock)

    current_model_file_name = "initial_model"
    torch.save(model.state_dict(), current_model_file_name)

    for epoch in range(1, args.epochs + 1):
        current_model_file_name = train_resnet_epoch(epoch, args, current_model_file_name)

        # Add the cross validation here
        accuracy = resnet_cross_validation(args, current_model_file_name)
        print("*** Cross validation accuracy = {} %".format(accuracy))

        # Decrease learning rate after every epoch
        args.lr = args.lr * 0.9
        print("        Decrease learning rate: {}".format(args.lr))


def train_resnet_epoch(model, epoch, train_seg_df, speaker_dict, args):
    """

    :param model:
    :param epoch:
    :param train_seg_df:
    :param args:
    :return:
    """
    device = torch.device("cuda:0")

    torch.manual_seed(args.seed)

    train_transform = []
    if not args.train_transformation == '':
        trans = args.train_transformation.split(',')
        for t in trans:
            if "CMVN" in t:
                train_transform.append(CMVN())
            if "FrequencyMask" in t:
                a = t.split(",")[0].split("(")[1]
                b = t.split(",")[1].split("(")[0]
                train_transform.append(FrequencyMask(a, b))
            if "TemporalMask" in t:
                a = t.split(",")[0].split("(")[1]
                train_transform.append(TemporalMask(a, b))
    train_set = VoxDataset(train_seg_df, speaker_dict, 500, transform=transforms.Compose(train_transform),
                           spec_aug_ratio=args.spec_aug, temp_aug_ratio=args.temp_aug)
    train_loader = DataLoader(train_set, batch_size=args.batch_size, shuffle=True, num_workers=15)

    optimizer = torch.optim.Adam(model.parameters())
    criterion = torch.nn.CrossEntropyLoss()

    accuracy = 0.0
    for batch_idx, (data, target, _, __) in enumerate(train_loader):
        target = target.squeeze()
        optimizer.zero_grad()
        output = model(data.to(device))
        loss = criterion(output, target.to(device))
        loss.backward()
        optimizer.step()
        accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()

        if batch_idx % args.log_interval == 0:
            logging.critical('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tAccuracy: {:.3f}'.format(
                epoch, batch_idx + 1, train_loader.__len__(),
                       100. * batch_idx / train_loader.__len__(), loss.item(),
                       100.0 * accuracy.item() / ((batch_idx + 1) * args.batch_size)))
    return model

def resnet_cross_validation(args, model, cv_seg_df, speaker_dict):
    """

    :param args:
    :param model:
    :param cv_seg_df:
    :return:
    """
    cv_transform = []
    if not args.cv_transformation == '':
        trans = args.cv_transformation.split(',')
        for t in trans:
            if "CMVN" in t:
                cv_transform.append(CMVN())
            if "FrequencyMask" in t:
                a = t.split(",")[0].split("(")[1]
                b = t.split(",")[1].split("(")[0]
                cv_transform.append(FrequencyMask(a, b))
            if "TemporalMask" in t:
                a = t.split(",")[0].split("(")[1]
                cv_transform.append(TemporalMask(a, b))
    cv_set = VoxDataset(cv_seg_df, speaker_dict, 500, transform=transforms.Compose(cv_transform),
                        spec_aug_ratio=args.spec_aug, temp_aug_ratio=args.temp_aug)
    cv_loader = DataLoader(cv_set, batch_size=args.batch_size, shuffle=False, num_workers=15)
    model.eval()
    device = torch.device("cuda:0")
    model.to(device)

    accuracy = 0.0
    print(cv_set.__len__())
    for batch_idx, (data, target, _, __) in enumerate(cv_loader):
        target = target.squeeze()
        output = model(data.to(device))
        accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()
    return 100. * accuracy.cpu().numpy() / ((batch_idx + 1) * args.batch_size)



# -*- coding: utf-8 -*-

# This package is a translation of a part of the BOSARIS toolkit.
# The authors thank Niko Brummer and Agnitio for allowing them to
# translate this code and provide the community with efficient structures
# and tools.
#
# The BOSARIS Toolkit is a collection of functions and classes in Matlab
# that can be used to calibrate, fuse and plot scores from speaker recognition
# (or other fields in which scores are used to test the hypothesis that two
# samples are from the same source) trials involving a model and a test segment.
# The toolkit was written at the BOSARIS2010 workshop which took place at the
# University of Technology in Brno, Czech Republic from 5 July to 6 August 2010.
# See the User Guide (available on the toolkit website)1 for a discussion of the
# theory behind the toolkit and descriptions of some of the algorithms used.
#
# The BOSARIS toolkit in MATLAB can be downloaded from `the website
# <https://sites.google.com/site/bosaristoolkit/>`_.

"""
This is the 'ndx' module
"""
import h5py
import logging
import numpy
import sys
from ..sidekit_wrappers import check_path_existance, deprecated

__author__ = "Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'
__credits__ = ["Niko Brummer", "Edward de Villiers"]


def diff(list1, list2):
    c = [item for item in list1 if item not in list2]
    c.sort()
    return c


def ismember(list1, list2):
    c = [item in list2 for item in list1]
    return c


class Ndx:
    """A class that encodes trial index information.  It has a list of
    model names and a list of test segment names and a matrix
    indicating which combinations of model and test segment are
    trials of interest.
    
    :attr modelset: list of unique models in a ndarray
    :attr segset:  list of unique test segments in a ndarray
    :attr trialmask: 2D ndarray of boolean. Rows correspond to the models 
            and columns to the test segments. True if the trial is of interest.
    """

    def __init__(self, ndx_file_name='',
                 models=numpy.array([]),
                 testsegs=numpy.array([])):
        """Initialize a Ndx object by loading information from a file
        in HDF5 or text format.

        :param ndx_file_name: name of the file to load
        """
        self.modelset = numpy.empty(0, dtype="|O")
        self.segset = numpy.empty(0, dtype="|O")
        self.trialmask = numpy.array([], dtype="bool")

        if ndx_file_name == '':
            modelset = numpy.unique(models)
            segset = numpy.unique(testsegs)
    
            trialmask = numpy.zeros((modelset.shape[0], segset.shape[0]), dtype="bool")
            for m in range(modelset.shape[0]):
                segs = testsegs[numpy.array(ismember(models, modelset[m]))]
                trialmask[m, ] = ismember(segset, segs)
    
            self.modelset = modelset
            self.segset = segset
            self.trialmask = trialmask
            assert self.validate(), "Wrong Ndx format"

        else:
            ndx = Ndx.read(ndx_file_name)
            self.modelset = ndx.modelset
            self.segset = ndx.segset
            self.trialmask = ndx.trialmask

    @check_path_existance
    def write(self, output_file_name):
        """ Save Ndx object in HDF5 format

        :param output_file_name: name of the file to write to
        """
        assert self.validate(), "Error: wrong Ndx format"

        with h5py.File(output_file_name, "w") as f:
            f.create_dataset("modelset", data=self.modelset.astype('S'),
                             maxshape=(None,),
                             compression="gzip",
                             fletcher32=True)
            f.create_dataset("segset", data=self.segset.astype('S'),
                             maxshape=(None,),
                             compression="gzip",
                             fletcher32=True)
            f.create_dataset("trial_mask", data=self.trialmask.astype('int8'),
                             maxshape=(None, None),
                             compression="gzip",
                             fletcher32=True)

    @check_path_existance
    def save_txt(self, output_file_name):

        """Save a Ndx object in a text file

        :param output_file_name: name of the file to write to
        """
        fid = open(output_file_name, 'w')
        for m in range(self.modelset.shape[0]):
            segs = self.segset[self.trialmask[m, ]]
            for s in segs:
                fid.write('{} {}\n'.format(self.modelset[m], s))
        fid.close()

    def filter(self, modlist, seglist, keep):
        """Removes some of the information in an Ndx. Useful for creating a
        gender specific Ndx from a pooled gender Ndx.  Depending on the
        value of \'keep\', the two input lists indicate the strings to
        retain or the strings to discard.

        :param modlist: a cell array of strings which will be compared with
                the modelset of 'inndx'.
        :param seglist: a cell array of strings which will be compared with
                the segset of 'inndx'.
        :param keep: a boolean indicating whether modlist and seglist are the
                models to keep or discard.

        :return: a filtered version of the current Ndx object.
        """
        if keep:
            keepmods = modlist
            keepsegs = seglist
        else:
            keepmods = diff(self.modelset, modlist)
            keepsegs = diff(self.segset, seglist)

        keepmodidx = numpy.array(ismember(self.modelset, keepmods))
        keepsegidx = numpy.array(ismember(self.segset, keepsegs))

        outndx = Ndx()
        outndx.modelset = self.modelset[keepmodidx]
        outndx.segset = self.segset[keepsegidx]
        tmp = self.trialmask[numpy.array(keepmodidx), :]
        outndx.trialmask = tmp[:, numpy.array(keepsegidx)]

        assert outndx.validate, "Wrong Ndx format"

        if self.modelset.shape[0] > outndx.modelset.shape[0]:
            logging.info('Number of models reduced from %d to %d', self.modelset.shape[0], outndx.modelset.shape[0])
        if self.segset.shape[0] > outndx.segset.shape[0]:
            logging.info('Number of test segments reduced from %d to %d', self.segset.shape[0], outndx.segset.shape[0])
        return outndx

    def validate(self):
        """Checks that an object of type Ndx obeys certain rules that
        must always be true.

        :return: a boolean value indicating whether the object is valid
        """
        ok = isinstance(self.modelset, numpy.ndarray)
        ok &= isinstance(self.segset, numpy.ndarray)
        ok &= isinstance(self.trialmask, numpy.ndarray)

        ok &= (self.modelset.ndim == 1)
        ok &= (self.segset.ndim == 1)
        ok &= (self.trialmask.ndim == 2)

        ok &= (self.trialmask.shape == (self.modelset.shape[0], self.segset.shape[0]))
        return ok

    @staticmethod
    def read(input_file_name):
        """Creates an Ndx object from the information in an hdf5 file.

        :param input_file_name: name of the file to read from
        """
        with h5py.File(input_file_name, "r") as f:
            ndx = Ndx()
            ndx.modelset = f.get("modelset")[()]
            ndx.segset = f.get("segset")[()]

            # if running python 3, need a conversion to unicode
            if sys.version_info[0] == 3:
                ndx.modelset = ndx.modelset.astype('U100', copy=False)
                ndx.segset = ndx.segset.astype('U100', copy=False)

            ndx.trialmask = f.get("trial_mask")[()].astype('bool')

            assert ndx.validate(), "Error: wrong Ndx format"
            return ndx

    @classmethod
    @check_path_existance
    def read_txt(cls, input_filename):
        """Creates an Ndx object from information stored in a text file.

        :param input_filename: name of the file to read from
        """
        ndx = Ndx()

        with open(input_filename, 'r') as fid:
            lines = [l.rstrip().split() for l in fid]

        models = numpy.empty(len(lines), '|O')
        testsegs = numpy.empty(len(lines), '|O')
        for ii in range(len(lines)):
            models[ii] = lines[ii][0]
            testsegs[ii] = lines[ii][1]

        modelset = numpy.unique(models)
        segset = numpy.unique(testsegs)

        trialmask = numpy.zeros((modelset.shape[0], segset.shape[0]), dtype="bool")
        for m in range(modelset.shape[0]):
            segs = testsegs[numpy.array(ismember(models, modelset[m]))]
            trialmask[m, ] = ismember(segset, segs)

        ndx.modelset = modelset
        ndx.segset = segset
        ndx.trialmask = trialmask

        assert ndx.validate(), "Wrong Ndx format"
        return ndx

    def merge(self, ndx_list):
        """Merges a list of Ndx objects into the current one.
        The resulting ndx must have all models and segment in the input
        ndxs (only once).  A trial in any ndx becomes a trial in the
        output ndx

        :param ndx_list: list of Ndx objects to merge
        """
        assert isinstance(ndx_list, list), "Input is not a list"
        for ndx in ndx_list:
            assert isinstance(ndx_list, list), \
                '{} {} {}'.format("Element ", ndx, " is not an Ndx")

        self.validate()
        for ndx2 in ndx_list:
            ndx_new = Ndx()
            ndx1 = self

            # create new ndx with empty masks
            ndx_new.modelset = numpy.union1d(ndx1.modelset, ndx2.modelset)
            ndx_new.segset = numpy.union1d(ndx1.segset, ndx2.segset)

            # expand ndx1 mask
            trials_1 = numpy.zeros((ndx_new.modelset.shape[0], ndx_new.segset.shape[0]), dtype="bool")
            model_index_a = numpy.argwhere(numpy.in1d(ndx_new.modelset, ndx1.modelset))
            model_index_b = numpy.argwhere(numpy.in1d(ndx1.modelset, ndx_new.modelset))
            seg_index_a = numpy.argwhere(numpy.in1d(ndx_new.segset, ndx1.segset))
            seg_index_b = numpy.argwhere(numpy.in1d(ndx1.segset, ndx_new.segset))
            trials_1[model_index_a[:, None], seg_index_a] = ndx1.trialmask[model_index_b[:, None], seg_index_b]

            # expand ndx2 mask
            trials_2 = numpy.zeros((ndx_new.modelset.shape[0], ndx_new.segset.shape[0]), dtype="bool")
            model_index_a = numpy.argwhere(numpy.in1d(ndx_new.modelset, ndx2.modelset))
            model_index_b = numpy.argwhere(numpy.in1d(ndx2.modelset, ndx_new.modelset))
            seg_index_a = numpy.argwhere(numpy.in1d(ndx_new.segset, ndx2.segset))
            seg_index_b = numpy.argwhere(numpy.in1d(ndx2.segset, ndx_new.segset))
            trials_2[model_index_a[:, None], seg_index_a] = ndx2.trialmask[model_index_b[:, None], seg_index_b]

            # merge masks
            trials = trials_1 | trials_2

            # build new ndx
            ndx_new.trialmask = trials
            self.modelset = ndx_new.modelset
            self.segset = ndx_new.segset
            self.trialmask = ndx_new.trialmask



# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2020 Anthony Larcher

"""

import h5py
import logging
import sys
import numpy
import torch
import torch.optim as optim
import torch.multiprocessing as mp
from collections import OrderedDict
from .xsets import XvectorMultiDataset, XvectorDataset, StatDataset
from ..bosaris import IdMap
from ..statserver import StatServer


__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2015-2020 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reS'











# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2020 Yevhenii Prokopalo, Anthony Larcher
"""

import logging
import numpy
import pickle
import torch
import torch.optim as optim
import torch.multiprocessing as mp
from torchvision import transforms
from collections import OrderedDict
from .xsets import XvectorMultiDataset, StatDataset, VoxDataset
from .xsets import FrequencyMask, CMVN, TemporalMask
from ..bosaris import IdMap
from ..statserver import StatServer
from torch.utils.data import DataLoader

__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2015-2020 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reS'


# logging.basicConfig(stream=sys.stdout, level=logging.INFO)


def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']


def split_file_list(batch_files, num_processes):
    # Cut the list of files into args.num_processes lists of files
    batch_sub_lists = [[]] * num_processes
    x = [ii for ii in range(len(batch_files))]
    for ii in range(num_processes):
        batch_sub_lists[ii - 1] = [batch_files[z + ii] for z in x[::num_processes] if (z + ii) < len(batch_files)]
    return batch_sub_lists


class Xtractor(torch.nn.Module):
    """
    Class that defines an x-vector extractor based on 5 convolutional layers and a mean standard deviation pooling
    """
    def __init__(self, spk_number, dropout):
        super(Xtractor, self).__init__()
        self.frame_conv0 = torch.nn.Conv1d(30, 512, 5, dilation=1)
        self.frame_conv1 = torch.nn.Conv1d(512, 512, 3, dilation=2)
        self.frame_conv2 = torch.nn.Conv1d(512, 512, 3, dilation=3)
        self.frame_conv3 = torch.nn.Conv1d(512, 512, 1)
        self.frame_conv4 = torch.nn.Conv1d(512, 3 * 512, 1)
        self.seg_lin0 = torch.nn.Linear(3 * 512 * 2, 512)
        self.dropout_lin0 = torch.nn.Dropout(p=dropout)
        self.seg_lin1 = torch.nn.Linear(512, 512)
        self.dropout_lin1 = torch.nn.Dropout(p=dropout)
        self.seg_lin2 = torch.nn.Linear(512, spk_number)
        #
        self.norm0 = torch.nn.BatchNorm1d(512)
        self.norm1 = torch.nn.BatchNorm1d(512)
        self.norm2 = torch.nn.BatchNorm1d(512)
        self.norm3 = torch.nn.BatchNorm1d(512)
        self.norm4 = torch.nn.BatchNorm1d(3 * 512)
        self.norm6 = torch.nn.BatchNorm1d(512)
        self.norm7 = torch.nn.BatchNorm1d(512)
        #
        self.activation = torch.nn.LeakyReLU(0.2)

    def produce_embeddings(self, x):
        """

        :param x:
        :return:
        """
        frame_emb_0 = self.norm0(self.activation(self.frame_conv0(x)))
        frame_emb_1 = self.norm1(self.activation(self.frame_conv1(frame_emb_0)))
        frame_emb_2 = self.norm2(self.activation(self.frame_conv2(frame_emb_1)))
        frame_emb_3 = self.norm3(self.activation(self.frame_conv3(frame_emb_2)))
        frame_emb_4 = self.norm4(self.activation(self.frame_conv4(frame_emb_3)))

        mean = torch.mean(frame_emb_4, dim=2)
        std = torch.std(frame_emb_4, dim=2)
        seg_emb = torch.cat([mean, std], dim=1)

        embedding_a = self.seg_lin0(seg_emb)
        return embedding_a

    def forward(self, x):
        """

        :param x:
        :return:
        """
        seg_emb_0 = self.produce_embeddings(x)
        # batch-normalisation after this layer
        seg_emb_1 = self.norm6(self.activation(seg_emb_0))
        # new layer with batch Normalization
        seg_emb_2 = self.norm7(self.activation(self.seg_lin1(self.dropout_lin1(seg_emb_1))))
        # No batch-normalisation after this layer
        result = self.activation(self.seg_lin2(seg_emb_2))
        return result

    def extract(self, x):
        """
        Extract x-vector given an input sequence of features

        :param x:
        :return:
        """
        embedding_a = self.produce_embeddings(x)
        embedding_b = self.seg_lin1(self.norm6(self.activation(embedding_a)))

        return embedding_a, embedding_b

    def init_weights(self):
        """
        Initialize the x-vector extract weights and biaises
        """
        torch.nn.init.normal_(self.frame_conv0.weight, mean=-0.5, std=0.1)
        torch.nn.init.normal_(self.frame_conv1.weight, mean=-0.5, std=0.1)
        torch.nn.init.normal_(self.frame_conv2.weight, mean=-0.5, std=0.1)
        torch.nn.init.normal_(self.frame_conv3.weight, mean=-0.5, std=0.1)
        torch.nn.init.normal_(self.frame_conv4.weight, mean=-0.5, std=0.1)
        torch.nn.init.xavier_uniform(self.seg_lin0.weight)
        torch.nn.init.xavier_uniform(self.seg_lin1.weight)
        torch.nn.init.xavier_uniform(self.seg_lin2.weight)

        torch.nn.init.constant(self.frame_conv0.bias, 0.1)
        torch.nn.init.constant(self.frame_conv1.bias, 0.1)
        torch.nn.init.constant(self.frame_conv2.bias, 0.1)
        torch.nn.init.constant(self.frame_conv3.bias, 0.1)
        torch.nn.init.constant(self.frame_conv4.bias, 0.1)
        torch.nn.init.constant(self.seg_lin0.bias, 0.1)
        torch.nn.init.constant(self.seg_lin1.bias, 0.1)
        torch.nn.init.constant(self.seg_lin2.bias, 0.1)


def xtrain(args):
    """
    Initialize and train an x-vector on a single GPU

    :param args:
    :return:
    """
    # If we start from an existing model
    if not args.init_model_name == '':
        # Load the model
        logging.critical("*** Load model from = {}/{}".format(args.model_path, args.init_model_name))
        model_file_name = '/'.join([args.model_path, args.init_model_name])
        model = torch.load(model_file_name)
        model.train()
    else:
        # Initialize a first model and save to disk
        model = Xtractor(args.class_number, args.dropout)
        model.train()

    if torch.cuda.device_count() > 1:
        print("Let's use", torch.cuda.device_count(), "GPUs!")
        model = torch.nn.DataParallel(model)

    model.cuda()

    # Split the training data in train and cv
    total_seg_df = pickle.load(open(args.batch_training_list, "rb"))

    speaker_dict = {}
    tmp = total_seg_df.speaker_id.unique()
    tmp.sort()
    for idx, spk in enumerate(tmp):
        speaker_dict[spk] = idx
    pickle.dump(speaker_dict, open("spk_dictionary.pkl", "wb"))

    cv_portion = 0.007
    idx = numpy.arange(len(total_seg_df))
    numpy.random.shuffle(idx)
    train_seg_df = total_seg_df.iloc[idx[:int((1 - cv_portion) * len(idx))]].reset_index()
    cv_seg_df = total_seg_df.iloc[idx[int((1 - cv_portion) * len(idx)):]].reset_index()

    current_model_file_name = "initial_model"
    torch.save(model.state_dict(), current_model_file_name)

    optimizer = torch.optim.SGD([
        {'params': model.module.frame_conv0.parameters(), 'weight_decay': args.l2_frame},
        {'params': model.module.frame_conv1.parameters(), 'weight_decay': args.l2_frame},
        {'params': model.module.frame_conv2.parameters(), 'weight_decay': args.l2_frame},
        {'params': model.module.frame_conv3.parameters(), 'weight_decay': args.l2_frame},
        {'params': model.module.frame_conv4.parameters(), 'weight_decay': args.l2_frame},
        {'params': model.module.seg_lin0.parameters(), 'weight_decay': args.l2_seg},
        {'params': model.module.seg_lin1.parameters(), 'weight_decay': args.l2_seg},
        {'params': model.module.seg_lin2.parameters(), 'weight_decay': args.l2_seg}
    ],
        lr=args.lr, momentum=0.9)

    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min')

    for epoch in range(1, args.epochs + 1):
        # Process one epoch and return the current model
        model = train_epoch(model, epoch, train_seg_df, speaker_dict, optimizer, args)

        # Add the cross validation here
        accuracy, val_loss = cross_validation(args, model, cv_seg_df, speaker_dict)
        logging.critical("*** Cross validation accuracy = {} %".format(accuracy))

        # Decrease learning rate according to the scheduler policy
        scheduler.step(val_loss)

        # return the file name of the new model
        base_name = "model"
        if not args.init_model_name == "":
            base_name = args.init_model_name
        current_model_file_name = "{}/{}_{}_epoch_{}".format(args.model_path, base_name, args.expe_id, epoch)
        torch.save(model, current_model_file_name)


def train_epoch(model, epoch, train_seg_df, speaker_dict, optimizer, args):
    """

    :param model:
    :param epoch:
    :param train_seg_df:
    :param speaker_dict:
    :param optimizer:
    :param args:
    :return:
    """
    device = torch.device("cuda:0")

    torch.manual_seed(args.seed)

    train_transform = []
    if not args.train_transformation == '':
        trans = args.train_transformation.split(',')
        for t in trans:
            if "CMVN" in t:
                train_transform.append(CMVN())
            if "FrequencyMask" in t:
                a = int(t.split("-")[0].split("(")[1])
                b = int(t.split("-")[1].split(")")[0])
                train_transform.append(FrequencyMask(a, b))
            if "TemporalMask" in t:
                a = int(t.split("(")[1].split(")")[0])
                train_transform.append(TemporalMask(a))
    train_set = VoxDataset(train_seg_df, speaker_dict, 500, transform=transforms.Compose(train_transform),
                           spec_aug_ratio=args.spec_aug, temp_aug_ratio=args.temp_aug)
    train_loader = DataLoader(train_set, batch_size=args.batch_size, shuffle=True, num_workers=15)

    criterion = torch.nn.CrossEntropyLoss()

    accuracy = 0.0
    for batch_idx, (data, target, _, __) in enumerate(train_loader):
        target = target.squeeze()
        optimizer.zero_grad()
        output = model(data.to(device))
        loss = criterion(output, target.to(device))
        loss.backward()
        optimizer.step()
        accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()

        if batch_idx % args.log_interval == 0:
            logging.critical('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tAccuracy: {:.3f}'.format(
                epoch, batch_idx + 1, train_loader.__len__(),
                       100. * batch_idx / train_loader.__len__(), loss.item(),
                       100.0 * accuracy.item() / ((batch_idx + 1) * args.batch_size)))
    return model


# def cross_validation(args, model):
#
#     with open(args.cross_validation_list, 'r') as fh:
#         cross_validation_list = [l.rstrip() for l in fh]
#     cv_loader = XvectorMultiDataset(cross_validation_list, args.batch_path)
#
#     model.eval()
#     device = torch.device("cuda:0")
#     model.to(device)
#
#     accuracy = 0.0
#     bi = 0
#     for batch_idx, (data, target) in enumerate(cv_loader):
#         output = model(data.to(device))
#         accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()
#         bi = batch_idx
#     return 100. * accuracy.cpu().numpy() / ((bi + 1) * args.batch_size)


def cross_validation(args, model, cv_seg_df, speaker_dict):
    """

    :param args:
    :param model:
    :param cv_seg_df:
    :return:
    """
    cv_transform = []
    if not args.cv_transformation == '':
        trans = args.cv_transformation.split(',')
        for t in trans:
            if "CMVN" in t:
                cv_transform.append(CMVN())
            if "FrequencyMask" in t:
                a = t.split(",")[0].split("(")[1]
                b = t.split(",")[1].split("(")[0]
                cv_transform.append(FrequencyMask(a, b))
            if "TemporalMask" in t:
                a = t.split(",")[0].split("(")[1]
                cv_transform.append(TemporalMask(a, b))
    cv_set = VoxDataset(cv_seg_df, speaker_dict, 500, transform=transforms.Compose(cv_transform),
                        spec_aug_ratio=args.spec_aug, temp_aug_ratio=args.temp_aug)
    cv_loader = DataLoader(cv_set, batch_size=args.batch_size, shuffle=False, num_workers=15)
    model.eval()
    device = torch.device("cuda:0")
    model.to(device)

    accuracy = 0.0
    criterion = torch.nn.CrossEntropyLoss()

    for batch_idx, (data, target, _, __) in enumerate(cv_loader):
        target = target.squeeze()
        output = model(data.to(device))
        accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()

    loss = criterion(output, target.to(device))

    return 100. * accuracy.cpu().numpy() / ((batch_idx + 1) * args.batch_size), loss


def xtrain_asynchronous(args):
    """
    Initialize and train an x-vector in asynchronous manner

    :param args:
    :return:
    """
    # Initialize a first model and save to disk
    model = Xtractor(args.class_number, args.dropout)
    current_model_file_name = "initial_model"
    torch.save(model.state_dict(), current_model_file_name)

    for epoch in range(1, args.epochs + 1):
        current_model_file_name = train_asynchronous_epoch(epoch, args, current_model_file_name)

        # Add the cross validation here
        accuracy = cross_asynchronous_validation(args, current_model_file_name)
        print("*** Cross validation accuracy = {} %".format(accuracy))

        # Decrease learning rate after every epoch
        args.lr = args.lr * 0.9
        print("        Decrease learning rate: {}".format(args.lr))


def train_asynchronous_epoch(epoch, args, initial_model_file_name):
    """
    Process one training epoch using an asynchronous implementation of the training

    :param epoch:
    :param args:
    :param initial_model_file_name:
    :return:
    """
    # Compute the megabatch number
    with open(args.batch_training_list, 'r') as fh:
        batch_file_list = [l.rstrip() for l in fh]

    # Shorten the batch_file_list to be a multiple of

    megabatch_number = len(batch_file_list) // (args.averaging_step * args.num_processes)
    megabatch_size = args.averaging_step * args.num_processes
    print("Epoch {}, number of megabatches = {}".format(epoch, megabatch_number))

    current_model = initial_model_file_name

    # For each sublist: run an asynchronous training and averaging of the model
    for ii in range(megabatch_number):
        print('Process megabatch [{} / {}]'.format(ii + 1, megabatch_number))
        current_model = train_asynchronous(epoch,
                                           args,
                                           current_model,
                                           batch_file_list[megabatch_size * ii: megabatch_size * (ii + 1)],
                                           ii,
                                           megabatch_number)  # function that split train, fuse and write the new model
    return current_model


def train_asynchronous(epoch, args, initial_model_file_name, batch_file_list, megabatch_idx, megabatch_number):
    """
    Process one mega-batch of data asynchronously, average the model parameters across
    subrocesses and return the updated version of the model

    :param epoch:
    :param args:
    :param initial_model_file_name:
    :param batch_file_list:
    :param megabatch_idx:
    :param megabatch_number:
    :return:
    """
    # Split the list of files for each process
    sub_lists = split_file_list(batch_file_list, args.num_processes)

    #
    output_queue = mp.Queue()
    # output_queue = multiprocessing.Queue()

    processes = []
    for rank in range(args.num_processes):
        p = mp.Process(target=train_asynchronous_worker,
                       args=(rank, epoch, args, initial_model_file_name, sub_lists[rank], output_queue)
                       )
        # We first train the model across `num_processes` processes
        p.start()
        processes.append(p)

    # Average the models and write the new one to disk
    asynchronous_model = []
    for ii in range(args.num_processes):
        asynchronous_model.append(dict(output_queue.get()))

    for p in processes:
        p.join()

    av_model = Xtractor(args.class_number, args.dropout)
    tmp = av_model.state_dict()

    average_param = dict()
    for k in list(asynchronous_model[0].keys()):
        average_param[k] = asynchronous_model[0][k]

        for mod in asynchronous_model[1:]:
            average_param[k] += mod[k]

        if 'num_batches_tracked' not in k:
            tmp[k] = torch.FloatTensor(average_param[k] / len(asynchronous_model))

    # return the file name of the new model
    current_model_file_name = "{}/model_{}_epoch_{}_batch_{}".format(args.model_path, args.expe_id, epoch,
                                                                     megabatch_idx)
    torch.save(tmp, current_model_file_name)
    if megabatch_idx == megabatch_number:
        torch.save(tmp, "{}/model_{}_epoch_{}".format(args.model_path, args.expe_id, epoch))

    return current_model_file_name


def train_asynchronous_worker(rank, epoch, args, initial_model_file_name, batch_list, output_queue):
    """


    :param rank:
    :param epoch:
    :param args:
    :param initial_model_file_name:
    :param batch_list:
    :param output_queue:
    :return:
    """
    model = Xtractor(args.class_number, args.dropout)
    model.load_state_dict(torch.load(initial_model_file_name))
    model.train()

    torch.manual_seed(args.seed + rank)
    train_loader = XvectorMultiDataset(batch_list, args.batch_path)

    device = torch.device("cuda:{}".format(rank))
    model.to(device)

    optimizer = optim.Adam([{'params': model.frame_conv0.parameters(), 'weight_decay': args.l2_frame},
                            {'params': model.frame_conv1.parameters(), 'weight_decay': args.l2_frame},
                            {'params': model.frame_conv2.parameters(), 'weight_decay': args.l2_frame},
                            {'params': model.frame_conv3.parameters(), 'weight_decay': args.l2_frame},
                            {'params': model.frame_conv4.parameters(), 'weight_decay': args.l2_frame},
                            {'params': model.seg_lin0.parameters(), 'weight_decay': args.l2_seg},
                            {'params': model.seg_lin1.parameters(), 'weight_decay': args.l2_seg},
                            {'params': model.seg_lin2.parameters(), 'weight_decay': args.l2_seg}
                            ], lr=args.lr)

    criterion = torch.nn.CrossEntropyLoss()

    accuracy = 0.0
    for batch_idx, (data, target) in enumerate(train_loader):
        optimizer.zero_grad()
        output = model(data.to(device))
        loss = criterion(output, target.to(device))
        loss.backward()
        optimizer.step()

        accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()

        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tAccuracy: {:.3f}'.format(
                epoch, batch_idx + 1, train_loader.__len__(),
                100. * batch_idx / train_loader.__len__(), loss.item(),
                100.0 * accuracy.item() / ((batch_idx + 1) * args.batch_size)))

    model_param = OrderedDict()
    params = model.state_dict()

    for k in list(params.keys()):
        model_param[k] = params[k].cpu().detach().numpy()
    output_queue.put(model_param)


def cross_asynchronous_validation(args, current_model_file_name):
    """

    :param args:
    :param current_model_file_name:
    :return:
    """
    with open(args.cross_validation_list, 'r') as fh:
        cross_validation_list = [l.rstrip() for l in fh]
        sub_lists = split_file_list(cross_validation_list, args.num_processes)

    #
    output_queue = mp.Queue()

    processes = []
    for rank in range(args.num_processes):
        p = mp.Process(target=cv_asynchronous_worker,
                       args=(rank, args, current_model_file_name, sub_lists[rank], output_queue)
                       )
        # We first evaluate the model across `num_processes` processes
        p.start()
        processes.append(p)

    # Average the models and write the new one to disk
    result = []
    for ii in range(args.num_processes):
        result.append(output_queue.get())

    for p in processes:
        p.join()

    # Compute the global accuracy
    accuracy = 0.0
    total_batch_number = 0
    for bn, acc in result:
        accuracy += acc
        total_batch_number += bn
    
    return 100. * accuracy / (total_batch_number * args.batch_size)


def cv_asynchronous_worker(rank, args, current_model_file_name, batch_list, output_queue):
    model = Xtractor(args.class_number, args.dropout)
    model.load_state_dict(torch.load(current_model_file_name))
    model.eval()

    cv_loader = XvectorMultiDataset(batch_list, args.batch_path)

    device = torch.device("cuda:{}".format(rank))
    model.to(device)

    accuracy = 0.0
    for batch_idx, (data, target) in enumerate(cv_loader):
        output = model(data.to(device))
        accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()
    output_queue.put((cv_loader.__len__(), accuracy.cpu().numpy()))


def extract_idmap(args, device_id, segment_indices, fs_params, idmap_name, output_queue):
    """
    Function that takes a model and an idmap and extract all x-vectors based on this model
    and return a StatServer containing the x-vectors
    """
    # device = torch.device("cuda:{}".format(device_ID))
    device = torch.device('cpu')

    # Create the dataset
    tmp_idmap = IdMap(idmap_name)
    idmap = IdMap()
    idmap.leftids = tmp_idmap.leftids[segment_indices]
    idmap.rightids = tmp_idmap.rightids[segment_indices]
    idmap.start = tmp_idmap.start[segment_indices]
    idmap.stop = tmp_idmap.stop[segment_indices]

    segment_loader = StatDataset(idmap, fs_params)

    # Load the model
    model_file_name = '/'.join([args.model_path, args.model_name])
    model = Xtractor(args.class_number, args.dropout)
    model.load_state_dict(torch.load(model_file_name))
    model.eval()

    # Get the size of embeddings
    emb_a_size = model.seg_lin0.weight.data.shape[0]
    emb_b_size = model.seg_lin1.weight.data.shape[0]

    # Create a Tensor to store all x-vectors on the GPU
    emb_1 = numpy.zeros((idmap.leftids.shape[0], emb_a_size)).astype(numpy.float32)
    emb_2 = numpy.zeros((idmap.leftids.shape[0], emb_b_size)).astype(numpy.float32)
    emb_3 = numpy.zeros((idmap.leftids.shape[0], emb_b_size)).astype(numpy.float32)
    emb_4 = numpy.zeros((idmap.leftids.shape[0], emb_b_size)).astype(numpy.float32)
    emb_5 = numpy.zeros((idmap.leftids.shape[0], emb_b_size)).astype(numpy.float32)
    emb_6 = numpy.zeros((idmap.leftids.shape[0], emb_b_size)).astype(numpy.float32)

    # Send on selected device
    model.to(device)

    # Loop to extract all x-vectors
    for idx, (model_id, segment_id, data) in enumerate(segment_loader):
        logging.critical('Process file {}, [{} / {}]'.format(segment_id, idx, segment_loader.__len__()))

        if list(data.shape)[2] < 20:
            pass
        else:
            seg_1, seg_2, seg_3, seg_4, seg_5, seg_6 = model.extract(data.to(device))
            emb_1[idx, :] = seg_1.detach().cpu()
            emb_2[idx, :] = seg_2.detach().cpu()
            emb_3[idx, :] = seg_3.detach().cpu()
            emb_4[idx, :] = seg_4.detach().cpu()
            emb_5[idx, :] = seg_5.detach().cpu()
            emb_6[idx, :] = seg_6.detach().cpu()

    output_queue.put((segment_indices, emb_1, emb_2, emb_3, emb_4, emb_5, emb_6))


def extract_parallel(args, fs_params):
    """

    :param args:
    :param fs_params:
    :return:
    """
    emb_a_size = 512
    emb_b_size = 512

    idmap = IdMap(args.idmap)

    x_server_1 = StatServer(idmap, 1, emb_a_size)
    x_server_2 = StatServer(idmap, 1, emb_b_size)
    x_server_3 = StatServer(idmap, 1, emb_b_size)
    x_server_4 = StatServer(idmap, 1, emb_b_size)
    x_server_5 = StatServer(idmap, 1, emb_b_size)
    x_server_6 = StatServer(idmap, 1, emb_b_size)

    x_server_1.stat0 = numpy.ones(x_server_1.stat0.shape)
    x_server_2.stat0 = numpy.ones(x_server_2.stat0.shape)
    x_server_3.stat0 = numpy.ones(x_server_3.stat0.shape)
    x_server_4.stat0 = numpy.ones(x_server_4.stat0.shape)
    x_server_5.stat0 = numpy.ones(x_server_5.stat0.shape)
    x_server_6.stat0 = numpy.ones(x_server_6.stat0.shape)

    # Split the indices
    mega_batch_size = idmap.leftids.shape[0] // args.num_processes

    logging.critical("Number of sessions to process: {}".format(idmap.leftids.shape[0]))

    segment_idx = []
    for ii in range(args.num_processes):
        segment_idx.append(
            numpy.arange(ii * mega_batch_size, numpy.min([(ii + 1) * mega_batch_size, idmap.leftids.shape[0]])))

    for idx, si in enumerate(segment_idx):
        logging.critical("Number of session on process {}: {}".format(idx, len(si)))

    # Extract x-vectors in parallel
    output_queue = mp.Queue()

    processes = []
    for rank in range(args.num_processes):
        p = mp.Process(target=extract_idmap,
                       args=(args, rank, segment_idx[rank], fs_params, args.idmap, output_queue)
                       )
        # We first train the model across `num_processes` processes
        p.start()
        processes.append(p)

    # Get the x-vectors and fill the StatServer
    for ii in range(args.num_processes):
        indices, seg_1, seg_2, seg_3, seg_4, seg_5, seg_6 = output_queue.get()
        x_server_1.stat1[indices, :] = seg_1
        x_server_2.stat1[indices, :] = seg_2
        x_server_3.stat1[indices, :] = seg_3
        x_server_4.stat1[indices, :] = seg_4
        x_server_5.stat1[indices, :] = seg_5
        x_server_6.stat1[indices, :] = seg_6

    for p in processes:
        p.join()

    return x_server_1, x_server_2, x_server_3, x_server_4, x_server_5, x_server_6


def extract_embeddings(args):
    """

    :param args:
    :param device_id:
    :param fs_params:
    :return:
    """
    device = torch.device("cuda:0")

    # Load the model
    logging.critical("*** Load model from = {}/{}".format(args.model_path, args.init_model_name))
    model_file_name = '/'.join([args.model_path, args.init_model_name])
    model = torch.load(model_file_name)
    model = torch.nn.DataParallel(model)
    model.eval()
    model.to(device)

    # Get the list of files
    total_seg_df = pickle.load(open(args.batch_training_list, "rb"))

    speaker_dict = {}
    tmp = total_seg_df.speaker_id.unique()
    tmp.sort()
    for idx, spk in enumerate(tmp):
        speaker_dict[spk] = idx

    extract_transform = [CMVN(), ]
    extract_set = VoxDataset(total_seg_df, speaker_dict, None, transform=transforms.Compose(extract_transform),
                             spec_aug_ratio=args.spec_aug, temp_aug_ratio=args.temp_aug)
    extract_loader = DataLoader(extract_set, batch_size=1, shuffle=False, num_workers=5)

    #CREER UN TENSEUR DE LA BONNE TAILLE POUR STOCKER LES X-VECTEURS

    for batch_idx, (data, target, _, __) in enumerate(extract_loader):
        print("extrait x-vecteur numero {}".format(batch_idx))
        embedding = model.produce_embeddings(data.to(device))
        #REMPLIR LE TENSEUR AVEC LE NOUVEAU X-VECTEUR

    #FAIRE CORRESPONDRE LES SPK_ID AVEC LES X-VECTEURS
    #RENVOYER LE TENSEUR DE X-VECTEURS SUR LE CPU OU L ECRTIRE SUR LE DISQUE



# -*- coding: utf-8 -*-
#
# This file is part of SIDEKIT.
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#
# SIDEKIT is a python package for speaker verification.
# Home page: http://www-lium.univ-lemans.fr/sidekit/
#    
# SIDEKIT is free software: you can redistribute it and/or modify
# it under the terms of the GNU LLesser General Public License as 
# published by the Free Software Foundation, either version 3 of the License, 
# or (at your option) any later version.
#
# SIDEKIT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Copyright 2014-2019 Anthony Larcher

:mod:`sidekit_io` provides methods to read and write from and to different 
formats.
"""

import h5py
import array
import numpy
import os
import pickle
import struct
import gzip
import logging
from sidekit.sidekit_wrappers import check_path_existance


__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2014-2019 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


def read_vect(filename):
    """Read vector in ALIZE binary format and return an array
    
    :param filename: name of the file to read from
    
    :return: a numpy.ndarray object
    """
    with open(filename, 'rb') as f:
        struct.unpack("<2l", f.read(8))
        data = array.array("d")
        data.fromstring(f.read())
    return numpy.array(data)


def read_matrix(filename):
    """Read matrix in ALIZE binary format and return a ndarray
    
    :param filename: name of the file to read from
    
    :return: a numpy.ndarray object
    """
    with open(filename, 'rb') as f:
        m_dim = struct.unpack("<2l", f.read(8))
        data = array.array("d")
        data.fromstring(f.read())
        T = numpy.array(data)
        T.resize(m_dim[0], m_dim[1])
    return T


@check_path_existance
def write_matrix(m, filename):
    """Write a  matrix in ALIZE binary format

    :param m: a 2-dimensional ndarray
    :param filename: name of the file to write in

    :exception: TypeError if m is not a 2-dimensional ndarray
    """
    if not m.ndim == 2:
        raise TypeError("To write vector, use write_vect")
    else:
        with open(filename, 'wb') as mf:
            data = numpy.array(m.flatten())
            mf.write(struct.pack("<l", m.shape[0]))
            mf.write(struct.pack("<l", m.shape[1]))
            mf.write(struct.pack("<" + "d" * m.shape[0] * m.shape[1], *data))


@check_path_existance
def write_vect(v, filename):
    """Write a  vector in ALIZE binary format

    :param v: a 1-dimensional ndarray
    :param filename: name of the file to write in
    
    :exception: TypeError if v is not a 1-dimensional ndarray
    """
    if not v.ndim == 1:
        raise TypeError("To write matrix, use write_matrix")
    else:
        with open(filename, 'wb') as mf:
            mf.write(struct.pack("<l", 1))
            mf.write(struct.pack("<l", v.shape[0]))
            mf.write(struct.pack("<" + "d" * v.shape[0], *v))


@check_path_existance
def write_matrix_int(m, filename):
    """Write matrix of int in ALIZE binary format
    
    :param m: a 2-dimensional ndarray of int
    :param filename: name of the file to write in
    """
    if not m.ndim == 2:
        raise TypeError("To write vector, use write_vect")
    if not m.dtype == 'int64':
        raise TypeError("m must be a ndarray of int64")
    with open(filename, 'wb') as mf:
        data = numpy.array(m.flatten())
        mf.write(struct.pack("<l", m.shape[0]))
        mf.write(struct.pack("<l", m.shape[1]))
        mf.write(struct.pack("<" + "l" * m.shape[0] * m.shape[1], *data))


def read_pickle(filename):
    """
    Read a generic pickle file and return the content

    :param filename: name of the pickle file to read

    :return: the content of the file
    """
    with gzip.open(filename, 'rb') as f:
        return pickle.load(f)


@check_path_existance
def write_pickle(obj, filename):
    """
    Dump an object in a picke file.

    :param obj: object to serialize and write
    :param filename: name of the file to write
    """
    if not (os.path.exists(os.path.dirname(filename)) or os.path.dirname(filename) == ''):
        os.makedirs(os.path.dirname(filename))
    with gzip.open(filename, 'wb') as f:
        pickle.dump(obj, f)


@check_path_existance
def write_tv_hdf5(data, output_filename):
    """
    Write the TotalVariability matrix, the mean and the residual covariance in HDF5 format.

    :param data: a tuple of three elements: the matrix, the mean vector and the inverse covariance vector
    :param output_filename: name fo the file to create
    """
    tv = data[0]
    tv_mean = data[1]
    tv_sigma = data[2]
    d = dict()
    d['tv/tv'] = tv
    d['tv/tv_mean'] = tv_mean
    d['tv/tv_sigma'] = tv_sigma
    write_dict_hdf5(d, output_filename)


def read_tv_hdf5(input_filename):
    """
    Read the TotalVariability matrix, the mean and the residual covariance from a HDF5 file.

    :param input_filename: name of the file to read from

    :return: a tuple of three elements: the matrix, the mean vector and the inverse covariance vector
    """
    with h5py.File(input_filename, "r") as f:
        tv = f.get("tv/tv")[()]
        tv_mean = f.get("tv/tv_mean")[()]
        tv_sigma = f.get("tv/tv_sigma")[()]
    return tv, tv_mean, tv_sigma


@check_path_existance
def write_dict_hdf5(data, output_filename):
    """
    Write a dictionary into a HDF5 file

    :param data: the dictionary to write
    :param output_filename: the name of the file to create
    """
    with h5py.File(output_filename, "w") as f:
        for key in data:
            value = data[key]
            if value is None:
                pass
            elif isinstance(value, numpy.ndarray) or isinstance(value, list):
                f.create_dataset(key,
                                 data=value,
                                 compression="gzip",
                                 fletcher32=True)
            else:
                f.create_dataset(key, data=value)


def read_key_hdf5(input_filename, key):
    """
    Read key value from a HDF5 file.

    :param input_filename: the name of the file to read from
    :param key: the name of the key

    :return: a value
    """
    with h5py.File(input_filename, "r") as f:
        return f.get(key)[()]


def read_dict_hdf5(input_filename):
    """
    Read a dictionary from an HDF5 file.

    :param input_filename: name of the file to read from

    :return: the dictionary
    """
    data = dict()
    with h5py.File(input_filename, "r") as f:
        for key in f.keys():
            logging.debug('key: '+key)
            for key2 in f.get(key).keys():
                data[key+'/'+key2] = f.get(key).get(key2)[()]
    return data


@check_path_existance
def write_norm_hdf5(data, output_filename):
    """
    Write the normalization parameters into a HDF5 file.

    :param data: a tuple of two lists. The first list contains mean vectors for each iteration,
    the second list contains covariance matrices for each iteration
    :param output_filename: name of the file to write in
    """
    with h5py.File(output_filename, "w") as f:
        means = data[0]
        covs = data[1]
        f.create_dataset("norm/means", data=means,
                         compression="gzip",
                         fletcher32=True)
        f.create_dataset("norm/covs", data=covs,
                         compression="gzip",
                         fletcher32=True)


def read_norm_hdf5(input_filename):
    """
    Read normalization parameters from a HDF5 file.

    :param input_filename: the name of the file to read from

    :return: a tuple of two lists. The first list contains mean vectors for each iteration,
        the second list contains covariance matrices for each iteration
    """
    with h5py.File(input_filename, "r") as f:
        means = f.get("norm/means")[()]
        covs = f.get("norm/covs")[()]
    return means, covs


@check_path_existance
def write_plda_hdf5(data, output_filename):
    """
    Write a PLDA model in a HDF5 file.

    :param data: a tuple of 4 elements: the mean vector, the between class covariance matrix,
        the within class covariance matrix and the residual matrix
    :param output_filename: the name of the file to read from
    """
    mean = data[0]
    mat_f = data[1]
    mat_g = data[2]
    sigma = data[3]
    with h5py.File(output_filename, "w") as f:
        f.create_dataset("plda/mean", data=mean,
                         compression="gzip",
                         fletcher32=True)
        f.create_dataset("plda/f", data=mat_f,
                         compression="gzip",
                         fletcher32=True)
        f.create_dataset("plda/g", data=mat_g,
                         compression="gzip",
                         fletcher32=True)
        f.create_dataset("plda/sigma", data=sigma,
                         compression="gzip",
                         fletcher32=True)


def read_plda_hdf5(input_filename):
    """
    Read a PLDA model from a HDF5 file.

    :param input_filename: the name of the file to read from

    :return: a tuple of 4 elements: the mean vector, the between class covariance matrix,
        the within class covariance matrix and the residual matrix
    """
    with h5py.File(input_filename, "r") as f:
        mean = f.get("plda/mean")[()]
        mat_f = f.get("plda/f")[()]
        mat_g = f.get("plda/g")[()]
        sigma = f.get("plda/sigma")[()]
    return mean, mat_f, mat_g, sigma


@check_path_existance
def write_fa_hdf5(data, output_filename):
    """
    Write a generic factor analysis model into a HDF5 file. (Used for instance for JFA storing)

    :param data: a tuple of 5 elements: the mean vector, the between class covariance matrix,
        the within class covariance matrix, the MAP matrix and the residual covariancematrix
    :param output_filename: the name of the file to write to
    :return:
    """
    mean = data[0]
    f = data[1]
    g = data[2]
    h = data[3]
    sigma = data[4]
    with h5py.File(output_filename, "w") as fh:
        kind = numpy.zeros(5, dtype="int16")  # FA with 5 matrix
        if mean is not None:
            kind[0] = 1
            fh.create_dataset("fa/mean", data=mean,
                              compression="gzip",
                              fletcher32=True)
        if f is not None:
            kind[1] = 1
            fh.create_dataset("fa/f", data=f,
                              compression="gzip",
                              fletcher32=True)
        if g is not None:
            kind[2] = 1
            fh.create_dataset("fa/g", data=g,
                              compression="gzip",
                              fletcher32=True)
        if h is not None:
            kind[3] = 1
            fh.create_dataset("fa/h", data=h,
                              compression="gzip",
                              fletcher32=True)
        if sigma is not None:
            kind[4] = 1
            fh.create_dataset("fa/sigma", data=sigma,
                              compression="gzip",
                              fletcher32=True)
        fh.create_dataset("fa/kind", data=kind,
                          compression="gzip",
                          fletcher32=True)


def read_fa_hdf5(input_filename):
    """
    Read a generic FA model from a HDF5 file

    :param input_filename: the name of the file to read from

    :return: a tuple of 5 elements: the mean vector, the between class covariance matrix,
        the within class covariance matrix, the MAP matrix and the residual covariancematrix
    """
    with h5py.File(input_filename, "r") as fh:
        kind = fh.get("fa/kind")[()]
        mean = f = g = h = sigma = None
        if kind[0] != 0:
            mean = fh.get("fa/mean")[()]
        if kind[1] != 0:
            f = fh.get("fa/f")[()]
        if kind[2] != 0:
            g = fh.get("fa/g")[()]
        if kind[3] != 0:
            h = fh.get("fa/h")[()]
        if kind[4] != 0:
            sigma = fh.get("fa/sigma")[()]
    return mean, f, g, h, sigma


def h5merge(output_filename, input_filename_list):
    """
    Merge a list of HDF5 files into a new one.

    :param output_filename: the name of the new file resulting from the merge.
    :param input_filename_list: list of thge input files
    """
    with h5py.File(output_filename, "w") as fo:
        for ifn in input_filename_list:
            logging.debug('read '+ifn)
            data = read_dict_hdf5(ifn)
            for key in data:
                value = data[key]
                if isinstance(value, numpy.ndarray) or isinstance(value, list):
                    fo.create_dataset(key,
                                      data=value,
                                      compression="gzip",
                                      fletcher32=True)
                else:
                    fo.create_dataset(key, data=value)


def init_logging(level=logging.INFO, filename=None):
    """
    Initialize a logger

    :param level: level of messages to catch
    :param filename: name of the output file
    """
    numpy.set_printoptions(linewidth=250, precision=4)
    frm = '%(asctime)s - %(levelname)s - %(message)s'
    root = logging.getLogger()
    if root.handlers:
        for handler in root.handlers:
            root.removeHandler(handler)
    logging.basicConfig(format=frm, level=level)
    if filename is not None:
        fh = logging.FileHandler(filename)
        fh.setFormatter(logging.Formatter(frm))
        fh.setLevel(level)
        root.addHandler(fh)


def write_matrix_hdf5(M, filename):
    with h5py.File(filename, "w") as h5f:
        h5f.create_dataset("matrix", data=M,
                           compression="gzip",
                           fletcher32=True)


def read_matrix_hdf5(filename):
    with h5py.File(filename, "r") as h5f:
        M = h5f.get("matrix")[()]
    return M
